/*
  SSHFS-MUX: SSHFS-MultipleXer
  Copyright (C) 2008 Nan Dun <dunnan@yl.is.s.u-tokyo.ac.jp>

  This program can be distributed under the terms of the GNU 
  GPL.
  See the file COPYING.
  
*/

#include "config.h"

#include <fuse.h>
#include <fuse_opt.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <semaphore.h>
#include <pthread.h>
#include <netdb.h>
#include <signal.h>
#include <pwd.h>
#include <limits.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <glib.h>

#include "cache.h"
#include "table.h"

#if FUSE_VERSION >= 23
#define SSHFSM_USE_INIT
#define SSHFSM_USE_DESTROY
#define SSHFSM_USE_CTRL
#endif

#define SSH_FXP_INIT                1
#define SSH_FXP_VERSION             2
#define SSH_FXP_OPEN                3
#define SSH_FXP_CLOSE               4
#define SSH_FXP_READ                5
#define SSH_FXP_WRITE               6
#define SSH_FXP_LSTAT               7
#define SSH_FXP_FSTAT               8
#define SSH_FXP_SETSTAT             9
#define SSH_FXP_FSETSTAT           10
#define SSH_FXP_OPENDIR            11
#define SSH_FXP_READDIR            12
#define SSH_FXP_REMOVE             13
#define SSH_FXP_MKDIR              14
#define SSH_FXP_RMDIR              15
#define SSH_FXP_REALPATH           16
#define SSH_FXP_STAT               17
#define SSH_FXP_RENAME             18
#define SSH_FXP_READLINK           19
#define SSH_FXP_SYMLINK            20
#define SSH_FXP_STATUS            101
#define SSH_FXP_HANDLE            102
#define SSH_FXP_DATA              103
#define SSH_FXP_NAME              104
#define SSH_FXP_ATTRS             105
#define SSH_FXP_EXTENDED          200
#define SSH_FXP_EXTENDED_REPLY    201

#define SSH_FILEXFER_ATTR_SIZE          0x00000001
#define SSH_FILEXFER_ATTR_UIDGID        0x00000002
#define SSH_FILEXFER_ATTR_PERMISSIONS   0x00000004
#define SSH_FILEXFER_ATTR_ACMODTIME     0x00000008
#define SSH_FILEXFER_ATTR_EXTENDED      0x80000000

#define SSH_FX_OK                            0
#define SSH_FX_EOF                           1
#define SSH_FX_NO_SUCH_FILE                  2
#define SSH_FX_PERMISSION_DENIED             3
#define SSH_FX_FAILURE                       4
#define SSH_FX_BAD_MESSAGE                   5
#define SSH_FX_NO_CONNECTION                 6
#define SSH_FX_CONNECTION_LOST               7
#define SSH_FX_OP_UNSUPPORTED                8

#define SSH_FXF_READ            0x00000001
#define SSH_FXF_WRITE           0x00000002
#define SSH_FXF_APPEND          0x00000004
#define SSH_FXF_CREAT           0x00000008
#define SSH_FXF_TRUNC           0x00000010
#define SSH_FXF_EXCL            0x00000020

#define PROTO_VERSION 3
#define MY_EOF 1
#define MAX_REPLY_LEN (1 << 17)
#define MAX_LINE 1024
#define RENAME_TEMP_CHARS 8
#define SFTP_SERVER_PATH "/usr/lib/sftp-server"
#define SSHNODELAY_SO "sshnodelay.so"
#define FUSERMOUNT "fusermount"

#define DEFAULT_CONF_PATH "~/.sshfsm"
#define DEFAULT_HOSTS_FILE_NAME "hosts"
#define DEFAULT_MAX_HOST_NUM 1024

#define COMMENT_PREFIX "//"
#define COMMENT_START_PREFIX "/*"
#define COMMENT_END_PREFIX "*/"

#ifdef SSHFSM_USE_CTRL
#define	CTRL_BUF_MAX_LEN 256
#define CTRL_MAX_CONN 1
#define CTRL_FILE_PATH "/tmp"
#endif

struct buffer {
	uint8_t *p;
	size_t len;
	size_t size;
};

struct list_head {
	struct list_head *prev;
	struct list_head *next;
};

struct request;
typedef void (*request_func)(struct request *);

struct request {
	unsigned int want_reply;
	sem_t ready;
	uint8_t reply_type;
	int replied;
	int error;
	struct buffer reply;
	struct timeval start;
	void *data;
	request_func end_func;
	size_t len;
	struct list_head list;
};

struct read_chunk {
	sem_t ready;
	off_t offset;
	size_t size;
	struct buffer data;
	int refs;
	int res;
	long modifver;
};

#ifdef SSHFSM_USE_CTRL
enum {
	CTRL_TYPE_UMOUNT,
	CTRL_TYPE_ADD,
	CTRL_TYPE_REMOVE,
	CTRL_TYPE_FINISH,
};

struct ctrl {
	int is_ctrl;
	int type;
	char *mountpoint;
};

static struct ctrl ctrl;
#endif

struct sshfsm_file {
	struct buffer handle;
	struct list_head write_reqs;
	pthread_cond_t write_finished;
	int write_error;
	struct read_chunk *readahead;
	off_t next_pos;
	int is_seq;
	int host_idx;
	int connver;
	int modifver;
	int refs;
};

struct sshfsm {
	char *directport;
	char *ssh_command;
	char *sftp_server;
	struct fuse_args ssh_args;
	char *workarounds;
	int rename_workaround;
	int nodelay_workaround;
	int nodelaysrv_workaround;
	int truncate_workaround;
	int buflimit_workaround;
	int transform_symlinks;
	int follow_symlinks;
	int no_check_root;
	int detect_uid;
	unsigned max_read;
	unsigned ssh_ver;
	int sync_write;
	int sync_read;
	int debug;
	int reconnect;
	int host_num_max;
	int	host_num;
	struct host **host_arr;
	int host_arr_empty_idx;
	pthread_mutex_t lock_host_arr;
	unsigned int randseed;
	unsigned local_uid;
	int remote_uid_detected;
	unsigned blksize;
	char *progname;
	unsigned outstanding_len;
	unsigned max_outstanding_len;
	pthread_cond_t outstanding_cond;
	char *conf_dir;
	char *hosts_file;
	char *mountpoint;
#ifdef	SSHFSM_USE_CTRL
	int ctrl_on;
	char *ctrl_file;
	int ctrl_fd;
	int ctrl_processing_thread_started;
#endif
};

static struct sshfsm sshfsm;

static const char *ssh_opts[] = {
	"AddressFamily",
	"BatchMode",
	"BindAddress",
	"ChallengeResponseAuthentication",
	"CheckHostIP",
	"Cipher",
	"Ciphers",
	"Compression",
	"CompressionLevel",
	"ConnectionAttempts",
	"ConnectTimeout",
	"GlobalKnownHostsFile",
	"GSSAPIAuthentication",
	"GSSAPIDelegateCredentials",
	"HostbasedAuthentication",
	"HostKeyAlgorithms",
	"HostKeyAlias",
	"HostName",
	"IdentityFile",
	"IdentitiesOnly",
	"LogLevel",
	"MACs",
	"NoHostAuthenticationForLocalhost",
	"NumberOfPasswordPrompts",
	"PasswordAuthentication",
	"Port",
	"PreferredAuthentications",
	"ProxyCommand",
	"PubkeyAuthentication",
	"RhostsRSAAuthentication",
	"RSAAuthentication",
	"ServerAliveInterval",
	"ServerAliveCountMax",
	"SmartcardDevice",
	"StrictHostKeyChecking",
	"TCPKeepAlive",
	"UsePrivilegedPort",
	"UserKnownHostsFile",
	"VerifyHostKeyDNS",
	NULL,
};

enum {
	KEY_PORT,
	KEY_COMPRESS,
	KEY_HELP,
	KEY_VERSION,
#ifdef SSHFSM_USE_CTRL
	KEY_CTRL_UMOUNT,
	KEY_CTRL_ADD,
	KEY_CTRL_REMOVE,
#endif
};

#define SSHFSM_OPT(t, p, v) { t, offsetof(struct sshfsm, p), v }

static struct fuse_opt sshfsm_opts[] = {
	SSHFSM_OPT("directport=%s",     directport, 0),
	SSHFSM_OPT("ssh_command=%s",    ssh_command, 0),
	SSHFSM_OPT("sftp_server=%s",    sftp_server, 0),
	SSHFSM_OPT("max_read=%u",       max_read, 0),
	SSHFSM_OPT("ssh_protocol=%u",   ssh_ver, 0),
	SSHFSM_OPT("-1",                ssh_ver, 1),
	SSHFSM_OPT("workaround=%s",     workarounds, 0),
	SSHFSM_OPT("idmap=none",        detect_uid, 0),
	SSHFSM_OPT("idmap=user",        detect_uid, 1),
	SSHFSM_OPT("sshfsm_sync",       sync_write, 1),
	SSHFSM_OPT("no_readahead",      sync_read, 1),
	SSHFSM_OPT("sshfsm_debug",      debug, 1),
	SSHFSM_OPT("reconnect",         reconnect, 1),
	SSHFSM_OPT("transform_symlinks", transform_symlinks, 1),
	SSHFSM_OPT("follow_symlinks",   follow_symlinks, 1),
	SSHFSM_OPT("no_check_root",     no_check_root, 1),
	SSHFSM_OPT("conf_dir=%s",       conf_dir, 0),
	SSHFSM_OPT("hosts_file=%s",     hosts_file, 0),
	SSHFSM_OPT("max_host=%d",       host_num_max, 0),
#ifdef SSHFSM_USE_CTRL
	SSHFSM_OPT("control=yes",     	ctrl_on, 1),
	SSHFSM_OPT("control=no",		ctrl_on, 0),
#endif

	FUSE_OPT_KEY("-p ",            KEY_PORT),
	FUSE_OPT_KEY("-C",             KEY_COMPRESS),
	FUSE_OPT_KEY("-V",             KEY_VERSION),
	FUSE_OPT_KEY("--version",      KEY_VERSION),
	FUSE_OPT_KEY("-h",             KEY_HELP),
	FUSE_OPT_KEY("--help",         KEY_HELP),
#ifdef SSHFSM_USE_CTRL
	FUSE_OPT_KEY("-u",			   KEY_CTRL_UMOUNT),
	FUSE_OPT_KEY("--umount",       KEY_CTRL_UMOUNT),
	FUSE_OPT_KEY("-a",			   KEY_CTRL_ADD),
	FUSE_OPT_KEY("--add",		   KEY_CTRL_ADD),
	FUSE_OPT_KEY("-r",			   KEY_CTRL_REMOVE),
	FUSE_OPT_KEY("--remove",	   KEY_CTRL_REMOVE),
#endif
	FUSE_OPT_END
};

static struct fuse_opt workaround_opts[] = {
	SSHFSM_OPT("none",       rename_workaround, 0),
	SSHFSM_OPT("none",       nodelay_workaround, 0),
	SSHFSM_OPT("none",       nodelaysrv_workaround, 0),
	SSHFSM_OPT("none",       truncate_workaround, 0),
	SSHFSM_OPT("none",       buflimit_workaround, 0),
	SSHFSM_OPT("all",        rename_workaround, 1),
	SSHFSM_OPT("all",        nodelay_workaround, 1),
	SSHFSM_OPT("all",        nodelaysrv_workaround, 1),
	SSHFSM_OPT("all",        truncate_workaround, 1),
	SSHFSM_OPT("all",        buflimit_workaround, 1),
	SSHFSM_OPT("rename",     rename_workaround, 1),
	SSHFSM_OPT("norename",   rename_workaround, 0),
	SSHFSM_OPT("nodelay",    nodelay_workaround, 1),
	SSHFSM_OPT("nonodelay",  nodelay_workaround, 0),
	SSHFSM_OPT("nodelaysrv", nodelaysrv_workaround, 1),
	SSHFSM_OPT("nonodelaysrv", nodelaysrv_workaround, 0),
	SSHFSM_OPT("truncate",   truncate_workaround, 1),
	SSHFSM_OPT("notruncate", truncate_workaround, 0),
	SSHFSM_OPT("buflimit",   buflimit_workaround, 1),
	SSHFSM_OPT("nobuflimit", buflimit_workaround, 0),
	FUSE_OPT_END
};

#define DEBUG(format, args...)						\
	do { if (sshfsm.debug) fprintf(stderr, format, args); } while(0)

static const char *type_name(uint8_t type)
{
	switch(type) {
	case SSH_FXP_INIT:           return "INIT";
	case SSH_FXP_VERSION:        return "VERSION";
	case SSH_FXP_OPEN:           return "OPEN";
	case SSH_FXP_CLOSE:          return "CLOSE";
	case SSH_FXP_READ:           return "READ";
	case SSH_FXP_WRITE:          return "WRITE";
	case SSH_FXP_LSTAT:          return "LSTAT";
	case SSH_FXP_FSTAT:          return "FSTAT";
	case SSH_FXP_SETSTAT:        return "SETSTAT";
	case SSH_FXP_FSETSTAT:       return "FSETSTAT";
	case SSH_FXP_OPENDIR:        return "OPENDIR";
	case SSH_FXP_READDIR:        return "READDIR";
	case SSH_FXP_REMOVE:         return "REMOVE";
	case SSH_FXP_MKDIR:          return "MKDIR";
	case SSH_FXP_RMDIR:          return "RMDIR";
	case SSH_FXP_REALPATH:       return "REALPATH";
	case SSH_FXP_STAT:           return "STAT";
	case SSH_FXP_RENAME:         return "RENAME";
	case SSH_FXP_READLINK:       return "READLINK";
	case SSH_FXP_SYMLINK:        return "SYMLINK";
	case SSH_FXP_STATUS:         return "STATUS";
	case SSH_FXP_HANDLE:         return "HANDLE";
	case SSH_FXP_DATA:           return "DATA";
	case SSH_FXP_NAME:           return "NAME";
	case SSH_FXP_ATTRS:          return "ATTRS";
	case SSH_FXP_EXTENDED:       return "EXTENDED";
	case SSH_FXP_EXTENDED_REPLY: return "EXTENDED_REPLY";
	default:                     return "???";
	}
}

#define container_of(ptr, type, member) ({				\
			const typeof( ((type *)0)->member ) *__mptr = (ptr); \
			(type *)( (char *)__mptr - offsetof(type,member) );})

#define list_entry(ptr, type, member)		\
	container_of(ptr, type, member)

static inline void * sshfsm_malloc(size_t size)
{	
	void *p;
	p = malloc(size);
	if (!p) {
		fprintf(stderr, "sshfsm: memory allocation failed\n");
		abort();
	}
	return p;
}

static inline void * sshfsm_realloc(void *p, size_t size)
{
	p = realloc(p, size);
	if (!p) {
		fprintf(stderr, "sshfsm: memory allocation failed\n");
       	abort();
	}
	return p;
}

static inline void * sshfsm_calloc(size_t cnt, size_t size)
{
	size_t s = cnt * size;
	void *p = malloc(s);
	if (!p) {
		fprintf(stderr, "sshfsm: memory allocation failed\n");
       	abort();
	}
	memset(p, 0, s);
	return p;
}

static inline char * sshfsm_strdup(const char *str)
{
	char *s = strdup(str);
	if (!s) {
		fprintf(stderr, "sshfsm: memory allocation failed\n");
       	abort();
	}
	return s;

}

static void list_init(struct list_head *head)
{
	head->next = head;
	head->prev = head;
}

static void list_add(struct list_head *new, struct list_head *head)
{
	struct list_head *prev = head;
	struct list_head *next = head->next;
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

static void list_del(struct list_head *entry)
{
	struct list_head *prev = entry->prev;
	struct list_head *next = entry->next;
	next->prev = prev;
	prev->next = next;
}

static int list_empty(const struct list_head *head)
{
	return head->next == head;
}

static inline void buf_init(struct buffer *buf, size_t size)
{
	if (size) {
		buf->p = (uint8_t *) malloc(size);
		if (!buf->p) {
			fprintf(stderr, 
					"sshfsm: memory allocation failed\n");
			abort();
		}
	} else
		buf->p = NULL;
	buf->len = 0;
	buf->size = size;
}

static inline void buf_free(struct buffer *buf)
{
	free(buf->p);
}

static inline void buf_finish(struct buffer *buf)
{
	buf->len = buf->size;
}

static inline void buf_clear(struct buffer *buf)
{
	buf_free(buf);
	buf_init(buf, 0);
}

static void buf_resize(struct buffer *buf, size_t len)
{
	buf->size = (buf->len + len + 63) & ~31;
	buf->p = (uint8_t *) realloc(buf->p, buf->size);
	if (!buf->p) {
		fprintf(stderr, "sshfsm: memory allocation failed\n");
		abort();
	}
}

static inline void buf_check_add(struct buffer *buf, size_t len)
{
	if (buf->len + len > buf->size)
		buf_resize(buf, len);
}

#define _buf_add_mem(b, d, l)			\
	buf_check_add(b, l);			\
	memcpy(b->p + b->len, d, l);		\
	b->len += l;


static inline void buf_add_mem(struct buffer *buf, const void *data,
                               size_t len)
{
	_buf_add_mem(buf, data, len);
}

static inline void buf_add_buf(struct buffer *buf, const struct buffer *bufa)
{
	_buf_add_mem(buf, bufa->p, bufa->len);
}

static inline void buf_add_uint8(struct buffer *buf, uint8_t val)
{
	_buf_add_mem(buf, &val, 1);
}

static inline void buf_add_uint32(struct buffer *buf, uint32_t val)
{
	uint32_t nval = htonl(val);
	_buf_add_mem(buf, &nval, 4);
}

static inline void buf_add_uint64(struct buffer *buf, uint64_t val)
{
	buf_add_uint32(buf, val >> 32);
	buf_add_uint32(buf, val & 0xffffffff);
}

static inline void buf_add_data(struct buffer *buf, const struct buffer *data)
{
	buf_add_uint32(buf, data->len);
	buf_add_mem(buf, data->p, data->len);
}

static inline void buf_add_string(struct buffer *buf, const char *str)
{
	struct buffer data;
	data.p = (uint8_t *) str;
	data.len = strlen(str);
	buf_add_data(buf, &data);
}

static inline void buf_add_path(struct buffer *buf, const int idx, 
								const char *path)
{
	char *realpath;
	
	realpath = g_strdup_printf("%s%s", sshfsm.host_arr[idx]->base_path,
				   			   path[1] ? path+1 : ".");
	buf_add_string(buf, realpath);
	g_free(realpath);
}

static int buf_check_get(struct buffer *buf, size_t len)
{
	if (buf->len + len > buf->size) {
		fprintf(stderr, "buffer too short\n");
		return -1;
	} else
		return 0;
}

static inline int buf_get_mem(struct buffer *buf, void *data, size_t len)
{
	if (buf_check_get(buf, len) == -1)
		return -1;
	memcpy(data, buf->p + buf->len, len);
	buf->len += len;
	return 0;
}

static inline int buf_get_uint8(struct buffer *buf, uint8_t *val)
{
	return buf_get_mem(buf, val, 1);
}

static inline int buf_get_uint32(struct buffer *buf, uint32_t *val)
{
	uint32_t nval;
	if (buf_get_mem(buf, &nval, 4) == -1)
		return -1;
	*val = ntohl(nval);
	return 0;
}

static inline int buf_get_uint64(struct buffer *buf, uint64_t *val)
{
	uint32_t val1;
	uint32_t val2;
	if (buf_get_uint32(buf, &val1) == -1 ||
	    buf_get_uint32(buf, &val2) == -1) {
		return -1;
	}
	*val = ((uint64_t) val1 << 32) + val2;
	return 0;
}

static inline int buf_get_data(struct buffer *buf, struct buffer *data)
{
	uint32_t len;
	if (buf_get_uint32(buf, &len) == -1 || len > buf->size - buf->len)
		return -1;
	buf_init(data, len + 1);
	data->size = len;
	if (buf_get_mem(buf, data->p, data->size) == -1) {
		buf_free(data);
		return -1;
	}
	return 0;
}

static inline int buf_get_string(struct buffer *buf, char **str)
{
	struct buffer data;
	if (buf_get_data(buf, &data) == -1)
		return -1;
	data.p[data.size] = '\0';
	*str = (char *) data.p;
	return 0;
}

static int buf_get_attrs(struct buffer *buf, struct stat *stbuf, int *flagsp)
{
	uint32_t flags;
	uint64_t size = 0;
	uint32_t uid = 0;
	uint32_t gid = 0;
	uint32_t atime = 0;
	uint32_t mtime = 0;
	uint32_t mode = S_IFREG | 0777;

	if (buf_get_uint32(buf, &flags) == -1)
		return -1;
	if (flagsp)
		*flagsp = flags;
	if ((flags & SSH_FILEXFER_ATTR_SIZE) &&
	    buf_get_uint64(buf, &size) == -1)
		return -1;
	if ((flags & SSH_FILEXFER_ATTR_UIDGID) &&
	    (buf_get_uint32(buf, &uid) == -1 ||
	     buf_get_uint32(buf, &gid) == -1))
		return -1;
	if ((flags & SSH_FILEXFER_ATTR_PERMISSIONS) &&
	    buf_get_uint32(buf, &mode) == -1)
		return -1;
	if ((flags & SSH_FILEXFER_ATTR_ACMODTIME)) {
		if (buf_get_uint32(buf, &atime) == -1 ||
		    buf_get_uint32(buf, &mtime) == -1)
			return -1;
	}
	if ((flags & SSH_FILEXFER_ATTR_EXTENDED)) {
		uint32_t extcount;
		unsigned i;
		if (buf_get_uint32(buf, &extcount) == -1)
			return -1;
		for (i = 0; i < extcount; i++) {
			struct buffer tmp;
			if (buf_get_data(buf, &tmp) == -1)
				return -1;
			buf_free(&tmp);
			if (buf_get_data(buf, &tmp) == -1)
				return -1;
			buf_free(&tmp);
		}
	}
	
	/* TODO: Properly set uid 
	if (sshfsm.remote_uid_detected && uid == sshfsm.remote_uid)
		uid = sshfsm.local_uid;
	*/

	memset(stbuf, 0, sizeof(struct stat));
	stbuf->st_mode = mode;
	stbuf->st_nlink = 1;
	stbuf->st_size = size;
	if (sshfsm.blksize) {
		stbuf->st_blksize = sshfsm.blksize;
		stbuf->st_blocks = ((size + sshfsm.blksize - 1) &
				    ~(sshfsm.blksize - 1)) >> 9;
	}
	stbuf->st_uid = uid;
	stbuf->st_gid = gid;
	stbuf->st_atime = atime;
	stbuf->st_ctime = stbuf->st_mtime = mtime;
	return 0;
}

struct dir_entry {
	char *name;
	struct stat stbuf;
};

void dir_entry_fill (GSList **lstp, const char *name, struct stat *stbuf)
{	
	struct dir_entry *entry = g_new(struct dir_entry, 1);
	entry->name = g_strdup(name);
	entry->stbuf = *stbuf;
	*lstp = g_slist_prepend(*lstp, (void *) entry);
}
typedef void (*dir_entry_fill_t) (GSList **lst, const char *name, 
								 struct stat *stbuf);

static int buf_get_entries(struct buffer *buf, GSList **lst,
                           dir_entry_fill_t filler)
{
	uint32_t count;
	unsigned i;

	if (buf_get_uint32(buf, &count) == -1)
		return -1;

	for (i = 0; i < count; i++) {
		int err = -1;
		char *name;
		char *longname;
		struct stat stbuf;
		if (buf_get_string(buf, &name) == -1)
			return -1;
		if (buf_get_string(buf, &longname) != -1) {
			free(longname);
			if (buf_get_attrs(buf, &stbuf, NULL) != -1) {
				if (sshfsm.follow_symlinks && S_ISLNK(stbuf.st_mode))
					stbuf.st_mode = 0;
				filler(lst, name, &stbuf);
				err = 0;
			}
		}
		free(name);
		if (err)
			return err;
	}
	return 0;
}

static void ssh_add_arg(const char *arg)
{
	if (fuse_opt_add_arg(&sshfsm.ssh_args, arg) == -1)
		_exit(1);
}


#ifdef SSH_NODELAY_WORKAROUND
static int do_ssh_nodelay_workaround(void)
{
	char *oldpreload = getenv("LD_PRELOAD");
	char *newpreload;
	char sopath[PATH_MAX];
	int res;

	snprintf(sopath, sizeof(sopath), "%s/%s", LIBDIR, SSHNODELAY_SO);
	res = access(sopath, R_OK);
	if (res == -1) {
		char *s;
		if (!realpath(sshfsm.progname, sopath))
			return -1;

		s = strrchr(sopath, '/');
		if (!s)
			s = sopath;
		else
			s++;

		if (s + strlen(SSHNODELAY_SO) >= sopath + sizeof(sopath))
			return -1;

		strcpy(s, SSHNODELAY_SO);
		res = access(sopath, R_OK);
		if (res == -1) {
			fprintf(stderr, "sshfsm: cannot find %s\n", SSHNODELAY_SO);
			return -1;
		}
	}

	newpreload = g_strdup_printf("%s%s%s", oldpreload ? oldpreload : "",
				     oldpreload ? " " : "", sopath);

	if (!newpreload || setenv("LD_PRELOAD", newpreload, 1) == -1) {
		fprintf(stderr, "warning: failed set LD_PRELOAD "
			"for ssh nodelay workaround\n");
	}
	g_free(newpreload);
	return 0;
}
#endif

static int start_ssh(int idx)
{
	int sockpair[2];
	int pid;
	struct host *hostp;
	
	hostp = sshfsm.host_arr[idx];
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair) == -1) {
		perror("failed to create socket pair");
		return -1;
	}
	hostp->fd = sockpair[0];

	pid = fork();
	if (pid == -1) {
		perror("failed to fork");
		return -1;
	} else if (pid == 0) {
		int devnull;

#ifdef SSH_NODELAY_WORKAROUND
		if (sshfsm.nodelay_workaround &&
		    do_ssh_nodelay_workaround() == -1) {
			fprintf(stderr, "warning: ssh nodelay workaround disabled\n");
		}
#endif

		if (sshfsm.nodelaysrv_workaround) {
			/*
			 * Hack to work around missing TCP_NODELAY
			 * setting in sshd
			 */
			sshfsm.ssh_args.argv[1] = "-X";
		}

		devnull = open("/dev/null", O_WRONLY);

		if (dup2(sockpair[1], 0) == -1 || dup2(sockpair[1], 1) == -1) {
			perror("failed to redirect input/output");
			_exit(1);
		}
		if (!sshfsm.debug && devnull != -1)
			dup2(devnull, 2);

		close(devnull);
		close(sockpair[0]);
		close(sockpair[1]);

		switch (fork()) {
			case -1:
				perror("failed to fork");
				_exit(1);
			case 0:
				break;
			default:
				_exit(0);
		}
		chdir("/");
		
		free(sshfsm.ssh_args.argv[2]);
		sshfsm.ssh_args.argv[2] = strdup(hostp->host_name);
		execvp(sshfsm.ssh_args.argv[0], sshfsm.ssh_args.argv);
		perror("execvp");
		_exit(1);
	}
	waitpid(pid, NULL, 0);
	close(sockpair[1]);
	return 0;
}

static int connect_to(int idx, char *port)
{
	int sock, opt, err;
	struct addrinfo *ai, hint;
	struct host *hostp;

	hostp = sshfsm.host_arr[idx];
	memset(&hint, 0, sizeof(hint));
	hint.ai_family = PF_INET;
	hint.ai_socktype = SOCK_STREAM;
	err = getaddrinfo(hostp->host_name, port, &hint, &ai);
	if (err) {
		fprintf(stderr, "failed to resolve %s:%s: %s\n", hostp->host_name, 
				port, gai_strerror(err));
		return -1;
	}
	sock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (sock == -1) {
		perror("failed to create socket");
		return -1;
	}
	err = connect(sock, ai->ai_addr, ai->ai_addrlen);
	if (err == -1) {
		perror("failed to connect");
		return -1;
	}
	opt = 1;
	err = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
	if (err == -1)
		perror("warning: failed to set TCP_NODELAY");

	freeaddrinfo(ai);

	hostp->fd = sock;
	return 0;
}

static int do_write(int idx, struct iovec *iov, size_t count)
{
	int res;
	struct host *hostp = sshfsm.host_arr[idx];
	
	while (count) {
		res = writev(hostp->fd, iov, count);
		if (res == -1) {
			perror("write");
			return -1;
		} else if (res == 0) {
			fprintf(stderr, "zero write\n");
			return -1;
		}
		do {
			if ((unsigned) res < iov->iov_len) {
				iov->iov_len -= res;
				iov->iov_base += res;
				break;
			} else {
				res -= iov->iov_len;
				count --;
				iov ++;
			}
		} while(count);
	}
	return 0;
}

static uint32_t sftp_get_id(int idx)
{
	return sshfsm.host_arr[idx]->idctr ++;
}

static void buf_to_iov(const struct buffer *buf, struct iovec *iov)
{
	iov->iov_base = buf->p;
	iov->iov_len = buf->len;
}

static size_t iov_length(const struct iovec *iov, unsigned long nr_segs)
{
	unsigned long seg;
	size_t ret = 0;

	for (seg = 0; seg < nr_segs; seg++)
		ret += iov[seg].iov_len;
	return ret;
}

#define SFTP_MAX_IOV 3

static int sftp_send_iov(int idx, uint8_t type, uint32_t id, 
						 struct iovec iov[], size_t count)
{
	int res;
	struct buffer buf;
	struct iovec iovout[SFTP_MAX_IOV];
	unsigned i;
	unsigned nout = 0;
	struct host *hostp;

	assert(count <= SFTP_MAX_IOV - 1);
	hostp = sshfsm.host_arr[idx];
	buf_init(&buf, 9);
	buf_add_uint32(&buf, iov_length(iov, count) + 5);
	buf_add_uint8(&buf, type);
	buf_add_uint32(&buf, id);
	buf_to_iov(&buf, &iovout[nout++]);
	for (i = 0; i < count; i++)
		iovout[nout++] = iov[i];
	pthread_mutex_lock(&hostp->lock_write);
	res = do_write(idx, iovout, nout);
	pthread_mutex_unlock(&hostp->lock_write);
	buf_free(&buf);
	return res;
}

static int do_read(int idx, struct buffer *buf)
{
	int res;
	uint8_t *p = buf->p;
	size_t size = buf->size;
	struct host *hostp = sshfsm.host_arr[idx];
	while (size) {
		res = read(hostp->fd, p, size);
		if (res == -1) {
			perror("read");
			return -1;
		} else if (res == 0) {
			fprintf(stderr, "remote host has disconnected\n");
			return -1;
		}
		size -= res;
		p += res;
	}
	return 0;
}

static int sftp_read(int idx, uint8_t *type, struct buffer *buf)
{
	int res;
	struct buffer buf2;
	uint32_t len;

	buf_init(&buf2, 5);
	res = do_read(idx, &buf2);
	if (res != -1) {
		if (buf_get_uint32(&buf2, &len) == -1)
			return -1;
		if (len > MAX_REPLY_LEN) {
			fprintf(stderr, "reply len too large: %u\n", len);
			return -1;
		}
		if (buf_get_uint8(&buf2, type) == -1)
			return -1;
		buf_init(buf, len - 1);
		res = do_read(idx, buf);
	}
	buf_free(&buf2);
	return res;
}

static void request_free(struct request *req)
{
	buf_free(&req->reply);
	sem_destroy(&req->ready);
	g_free(req);
}

static void chunk_free(struct read_chunk *chunk)
{
	buf_free(&chunk->data);
	sem_destroy(&chunk->ready);
	g_free(chunk);
}

static void chunk_put(struct read_chunk *chunk)
{
	if (chunk) {
		chunk->refs--;
		if (!chunk->refs)
			chunk_free(chunk);
	}
}

static void chunk_put_locked(int idx, struct read_chunk *chunk)
{
	pthread_mutex_lock(&sshfsm.host_arr[idx]->lock);
	chunk_put(chunk);
	pthread_mutex_unlock(&sshfsm.host_arr[idx]->lock);
}

static int clean_req(void *key_, struct request *req)
{
	(void) key_;

	req->error = -EIO;
	if (req->want_reply)
		sem_post(&req->ready);
	else {
		if (req->end_func)
			req->end_func(req);
		request_free(req);
	}
	return TRUE;
}

static int process_one_request(int idx)
{
	int res;
	struct buffer buf;
	uint8_t type;
	struct request *req;
	uint32_t id;
	struct host *hostp = sshfsm.host_arr[idx];

	buf_init(&buf, 0);
	res = sftp_read(idx, &type, &buf);
	if (res == -1)
		return -1;
	if (buf_get_uint32(&buf, &id) == -1)
		return -1;

	pthread_mutex_lock(&hostp->lock);
	req = (struct request *)
		g_hash_table_lookup(hostp->reqtab, GUINT_TO_POINTER(id));
	if (req == NULL)
		fprintf(stderr, "request %i not found\n", id);
	else {
		int was_over;

		was_over = sshfsm.outstanding_len > sshfsm.max_outstanding_len;
		sshfsm.outstanding_len -= req->len;
		if (was_over && sshfsm.outstanding_len <= 
			sshfsm.max_outstanding_len) {
			pthread_cond_broadcast(&sshfsm.outstanding_cond);
		}
		g_hash_table_remove(hostp->reqtab, GUINT_TO_POINTER(id));
	}
	pthread_mutex_unlock(&hostp->lock);
	if (req != NULL) {
		if (sshfsm.debug) {
			struct timeval now;
			unsigned int difftime;
			gettimeofday(&now, NULL);
			difftime = (now.tv_sec - req->start.tv_sec) * 1000;
			difftime += (now.tv_usec - req->start.tv_usec) / 1000;
			DEBUG("  [%05i] %14s %8ubytes (%ims)\n", id,
			      type_name(type),
			      (unsigned) buf.size + 5, difftime);
		}
		req->reply = buf;
		req->reply_type = type;
		req->replied = 1;
		if (req->want_reply)
			sem_post(&req->ready);
		else {
			if (req->end_func) {
				pthread_mutex_lock(&hostp->lock);
				req->end_func(req);
				pthread_mutex_unlock(&hostp->lock);
			}
			request_free(req);
		}
	} else
		buf_free(&buf);

	return 0;
}

static void *process_requests(void *data)
{	
	int idx, *idxp;
	struct host *hostp;
	
	idxp = (int *) data;
	idx = *idxp;
	hostp = sshfsm.host_arr[idx];
	while (1) {
		if (process_one_request(idx) == -1)
			break;
	}

	if (!sshfsm.reconnect) {
		/* harakiri */
		kill(getpid(), SIGTERM);
	} else {
		pthread_mutex_lock(&hostp->lock);
		hostp->processing_thread_started = 0;
		close(hostp->fd);
		hostp->fd = -1;
		g_hash_table_foreach_remove(hostp->reqtab, (GHRFunc) clean_req,
					    NULL);
		hostp->connver ++;
		pthread_mutex_unlock(&hostp->lock);
	}

	free(data);
	return NULL;
}

static int sftp_init_reply_ok(int idx, struct buffer *buf, uint32_t *version)
{
	uint32_t len;
	uint8_t type;

	if (buf_get_uint32(buf, &len) == -1)
		return -1;

	if (len < 5 || len > MAX_REPLY_LEN)
		return 1;

	if (buf_get_uint8(buf, &type) == -1)
		return -1;

	if (type != SSH_FXP_VERSION)
		return 1;

	if (buf_get_uint32(buf, version) == -1)
		return -1;

	if (len > 5) {
		struct buffer buf2;
		buf_init(&buf2, len - 5);
		return do_read(idx, &buf2);
	}
	return 0;
}

static int sftp_find_init_reply(int idx, uint32_t *version)
{
	int res;
	struct buffer buf;

	buf_init(&buf, 9);
	res = do_read(idx, &buf);
	while (res != -1) {
		struct buffer buf2;

		res = sftp_init_reply_ok(idx, &buf, version);
		if (res <= 0)
			break;

		/* Iterate over any rubbish until the version reply is found */
		DEBUG("%c", *buf.p);
		memmove(buf.p, buf.p + 1, buf.size - 1);
		buf.len = 0;
		buf2.p = buf.p + buf.size - 1;
		buf2.size = 1;
		res = do_read(idx, &buf2);
	}
	buf_free(&buf);
	return res;
}

static int sftp_init(int idx)
{
	int res = -1;
	uint32_t version = 0;
	struct buffer buf;

	buf_init(&buf, 0);
	if (sftp_send_iov(idx, SSH_FXP_INIT, PROTO_VERSION, NULL, 0) == -1)
		goto out;
	if (sftp_find_init_reply(idx, &version) == -1)
		goto out;

	sshfsm.host_arr[idx]->server_version = version;
	DEBUG("Host[%d] server version: %i\n", idx, 
	      sshfsm.host_arr[idx]->server_version);
	if (version > PROTO_VERSION) {
		fprintf(stderr,
			"Warning: server uses version: %i, we support: %i\n",
			version, PROTO_VERSION);
	}
	res = 0;

out:
	buf_free(&buf);
	return res;
}

static int sftp_error_to_errno(uint32_t error)
{
	switch (error) {
	case SSH_FX_OK:                return 0;
	case SSH_FX_NO_SUCH_FILE:      return ENOENT;
	case SSH_FX_PERMISSION_DENIED: return EACCES;
	case SSH_FX_FAILURE:           return EPERM;
	case SSH_FX_BAD_MESSAGE:       return EBADMSG;
	case SSH_FX_NO_CONNECTION:     return ENOTCONN;
	case SSH_FX_CONNECTION_LOST:   return ECONNABORTED;
	case SSH_FX_OP_UNSUPPORTED:    return EOPNOTSUPP;
	default:                       return EIO;
	}
}

static void sftp_detect_uid(int idx)
{
	int flags;
	uint32_t id = sftp_get_id(idx);
	uint32_t replid;
	uint8_t type;
	struct buffer buf;
	struct stat stbuf;
	struct iovec iov[1];

	buf_init(&buf, 5);
	buf_add_string(&buf, ".");
	buf_to_iov(&buf, &iov[0]);
	if (sftp_send_iov(idx, SSH_FXP_STAT, id, iov, 1) == -1)
		goto out;
	buf_clear(&buf);
	if (sftp_read(idx, &type, &buf) == -1)
		goto out;
	if (type != SSH_FXP_ATTRS && type != SSH_FXP_STATUS) {
		fprintf(stderr, "protocol error\n");
		goto out;
	}
	if (buf_get_uint32(&buf, &replid) == -1)
		goto out;
	if (replid != id) {
		fprintf(stderr, "bad reply ID\n");
		goto out;
	}
	if (type == SSH_FXP_STATUS) {
		uint32_t serr;
		if (buf_get_uint32(&buf, &serr) == -1)
			goto out;

		fprintf(stderr, "failed to stat home directory (%i)\n", serr);
		goto out;
	}
	if (buf_get_attrs(&buf, &stbuf, &flags) == -1)
		goto out;

	if (!(flags & SSH_FILEXFER_ATTR_UIDGID))
		goto out;

	sshfsm.host_arr[idx]->uid = stbuf.st_uid;

  out:
	if (!sshfsm.remote_uid_detected)
		fprintf(stderr, "failed to detect remote user ID\n");

	buf_free(&buf);
}

static void sftp_detect_uid_all() 
{
	int i;
	
	/* TODO: Thread */
	for (i = 0; i < sshfsm.host_num; i++)
		sftp_detect_uid(i);

	/*TODO: Define uid selection rule */
	sshfsm.local_uid = getuid();
	sshfsm.remote_uid_detected = 1;
}

static int sftp_check_root(int idx)
{
	int flags;
	uint32_t id = sftp_get_id(idx);
	uint32_t replid;
	uint8_t type;
	struct buffer buf;
	struct stat stbuf;
	struct iovec iov[1];
	int err = -1;
	struct host *hostp = sshfsm.host_arr[idx];
	const char *remote_dir = hostp->base_path[0] ? hostp->base_path : ".";

	buf_init(&buf, 0);
	buf_add_string(&buf, remote_dir);
	buf_to_iov(&buf, &iov[0]);
	if (sftp_send_iov(idx, SSH_FXP_STAT, id, iov, 1) == -1)
		goto out;
	buf_clear(&buf);
	if (sftp_read(idx, &type, &buf) == -1)
		goto out;
	if (type != SSH_FXP_ATTRS && type != SSH_FXP_STATUS) {
		fprintf(stderr, "protocol error\n");
		goto out;
	}
	if (buf_get_uint32(&buf, &replid) == -1)
		goto out;
	if (replid != id) {
		fprintf(stderr, "bad reply ID\n");
		goto out;
	}
	if (type == SSH_FXP_STATUS) {
		uint32_t serr;
		if (buf_get_uint32(&buf, &serr) == -1)
			goto out;
		
		fprintf(stderr, "%s:%s: %s\n", 
				hostp->host_name, remote_dir, 
				strerror(sftp_error_to_errno(serr)));

		goto out;
	}
	if (buf_get_attrs(&buf, &stbuf, &flags) == -1)
		goto out;

	if (!(flags & SSH_FILEXFER_ATTR_PERMISSIONS))
		goto out;

	if (!S_ISDIR(stbuf.st_mode)) {
		fprintf(stderr, "%s:%s: Not a directory\n", hostp->host_name,
			remote_dir);
		goto out;
	
	}

	err = 0;

out:
	buf_free(&buf);
	return err;
}

static void * sftp_check_root_thread(void *data)
{
	int *idxp, err;
	
	idxp = (int *) data;
	err = sftp_check_root(*idxp);
	if (err)
		pthread_exit((void *) -1);
	
	pthread_exit((void *) 0);
}

static int sftp_check_root_all() 
{
	int err, i, *idxp, retval, ret, flag = 0;
	pthread_t *thread;
	pthread_attr_t attr;
	
	idxp = (int *) sshfsm_malloc(sizeof(int) * sshfsm.host_num);
	thread = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * sshfsm.host_num);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	err = 0;
	for (i = 0; i < sshfsm.host_num; i++) {
		idxp[i] = i;
		err = pthread_create(&thread[i], &attr, sftp_check_root_thread, 
							 &idxp[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
		
	for (i = 0; i < sshfsm.host_num; i++) {
		err = pthread_join(thread[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", strerror(err));
			goto out;
		}
		if (retval == -1) 
			goto out;
	}
	
	flag = 1;
 out:
	pthread_attr_destroy(&attr);
	free(idxp);
	free(thread);
	ret = flag ? 0 : -1;
	return ret;
}

static int connect_remote(int idx)
{
	int err;

	if (sshfsm.directport)
		err = connect_to(idx, sshfsm.directport);
	else
		err = start_ssh(idx);
	
	if (!err)
		err = sftp_init(idx);

	return err;
}

static void * connect_remote_thread(void *data)
{
	int *idxp = (int *) data;
	int err = 0;
	err = connect_remote(*idxp);
	err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int connect_remote_all() {
	int err, i, *idxp, retval, ret, flag = 0;
	pthread_t *threads;
	pthread_attr_t attr;
	
	idxp = (int *) sshfsm_malloc(sizeof(int) * sshfsm.host_num);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * 
										  sshfsm.host_num);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	err = 0;
	for (i = 0; i < sshfsm.host_num; i++) {
		idxp[i] = i;
		err = pthread_create(&threads[i], &attr, connect_remote_thread, 
							 &idxp[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < sshfsm.host_num; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		if (retval == 0)
			flag = 1;
	}

 out:
	pthread_attr_destroy(&attr);
	free(idxp);
	free(threads);
	ret = flag ? 0 : -1;
	return ret;
}

static int start_processing_thread(int idx)
{
	int err;
	pthread_t thread_id;
	sigset_t oldset;
	sigset_t newset;
	int *datap;
	struct host *hostp;

	hostp = sshfsm.host_arr[idx];
	if (hostp->processing_thread_started)
		return 0;
	
	if (hostp->fd == -1) {
		err = connect_remote(idx);
		if (err)
			return -EIO;
	}

	datap = (int *) sshfsm_malloc(sizeof(int));
	*datap = idx;
	sigemptyset(&newset); sigaddset(&newset, SIGTERM);
	sigaddset(&newset, SIGINT);
	sigaddset(&newset, SIGHUP);
	sigaddset(&newset, SIGQUIT);
	pthread_sigmask(SIG_BLOCK, &newset, &oldset);
	err = pthread_create(&thread_id, NULL, process_requests, (void *) datap);
	if (err) {
		fprintf(stderr, "failed to create thread: %s\n", strerror(err));
		return -EIO;
	}
	pthread_detach(thread_id);
	pthread_sigmask(SIG_SETMASK, &oldset, NULL);
	hostp->processing_thread_started = 1;
	return 0;
}

static int start_processing_thread_all()
{
	int err, i;

	err = 0;
	for (i = 0; i < sshfsm.host_num; i++)
		err = start_processing_thread(i);	

	return err;
}

static int host_do_join(const char *hoststr)
{
	int res = table_host_map_lookup(hoststr);
	if (res != -1)
		return EEXIST;
	
	struct host *hostp;
	struct table_entry *entry = NULL;
	char *tmp, *cp, *s;
	const char delimiters[] = ":";
	int idx, i, j;
	
	assert(sshfsm.host_num + 1 <= sshfsm.host_num_max);
	idx = sshfsm.host_arr_empty_idx;
	sshfsm.host_arr[idx] = hostp = (struct host *)
								   sshfsm_malloc(sizeof(struct host));
	cp = strdup(hoststr);
	tmp = strtok(cp, delimiters);
	if (!tmp) {
		res = EIO;
		goto out;
	} else 
		hostp->host_name = strdup(tmp);

	tmp = strtok(NULL, delimiters);
	if (!tmp)
		hostp->base_path = strdup("");
	else {
		if (tmp[strlen(tmp) - 1] != '/') {
			s = (char *) sshfsm_calloc(strlen(tmp) + 2, sizeof(char));
			snprintf(s, strlen(tmp)+2, "%s/", tmp);
			hostp->base_path = s;
		} else
			hostp->base_path = strdup(tmp);
	}

	pthread_mutex_init(&hostp->lock, NULL);
	hostp->reqtab = g_hash_table_new(NULL, NULL);
	if (!hostp->reqtab) {
		fprintf(stderr, "failed to create request table\n");
		res = EIO;
		goto out;
	}
	pthread_mutex_init(&(hostp->lock_write), NULL);
	hostp->connver = hostp->processing_thread_started = 0;
	hostp->rank = idx;
	hostp->fd = -1;
	
	if (connect_remote(idx) == -1) {	
		res = EIO;
		goto out;
	}
	if (sshfsm.detect_uid)
		sftp_detect_uid(idx);
 	
	if (!sshfsm.no_check_root && sftp_check_root(idx) == -1) {
		res = ENOENT;
		goto out;
	}

	start_processing_thread(idx);
	table_host_map_insert(hoststr, idx);
	sshfsm.host_num ++;
	entry = table_entry_new(sshfsm.host_num);
	for (i = 0, j = 0; i < sshfsm.host_num && j < sshfsm.host_num_max; j++)
		if (sshfsm.host_arr[j] != NULL)
			entry->idx_arr[i++] = j;

	entry->size = sshfsm.host_num;
	entry->atime = time(NULL);
	entry->checksum = table_entry_hash(entry);
	table_insert("/", entry);
	cache_invalidate_all();
	table_tlb_purge_all();
	do {
		sshfsm.host_arr_empty_idx = (sshfsm.host_arr_empty_idx + 1) %
									sshfsm.host_num_max;
	} while (sshfsm.host_arr[sshfsm.host_arr_empty_idx] != NULL);
	res = 0;
  out:
  	free(cp);
	return res;
}

static int host_join(const char *hoststr)
{
	pthread_mutex_lock(&sshfsm.lock_host_arr);
	int ret = host_do_join(hoststr);
	pthread_mutex_unlock(&sshfsm.lock_host_arr);
	return ret;
}

static int host_do_leave(const char *hoststr)
{	
	int i, j;
	struct table_entry *entry = NULL;
	int idx = table_host_map_lookup(hoststr);
	if (idx < 0)
		return ENOENT;

	struct host *hostp = sshfsm.host_arr[idx];
	table_host_map_remove(hoststr);
	sshfsm.host_num --;
	sshfsm.host_arr[idx] = NULL;
	entry = table_entry_new(sshfsm.host_num);
	for (i = 0, j = 0; i < sshfsm.host_num && j < sshfsm.host_num_max; j++)
		if (sshfsm.host_arr[j] != NULL)
			entry->idx_arr[i++] = j;
	entry->size = sshfsm.host_num;
	entry->atime = time(NULL);
	entry->checksum = table_entry_hash(entry);
	table_insert("/", entry);
	cache_invalidate_all();
	table_tlb_purge_all();
	sshfsm.host_arr_empty_idx = idx;

	close(hostp->fd);
	g_hash_table_destroy(hostp->reqtab);
	pthread_mutex_destroy(&hostp->lock);
	pthread_mutex_destroy(&hostp->lock_write);
	free(hostp->host_name);
	free(hostp->base_path);
	return 0;
}

static int host_leave(const char *hoststr)
{
	pthread_mutex_lock(&sshfsm.lock_host_arr);
	int ret = host_do_leave(hoststr);
	pthread_mutex_unlock(&sshfsm.lock_host_arr);
	return ret;
}

static int parse_hosts_file(const char *file)
{
	FILE *fp;
	char *filename, *buf, *cp, *tmp, *s;
	const char delimiters[] = ":";
	struct host *hostp;
	int is_comment = 0;

	filename = g_strdup_printf("%s/%s", sshfsm.conf_dir, file);
	fp = fopen(filename, "r");
	if (!fp) {
		fprintf(stderr, "open file %s failed: %s\n", filename,
				strerror(errno));
		g_free(filename);
		return -1;
	}
	buf = sshfsm_calloc(MAX_LINE, sizeof(char));
	while (fgets(buf, MAX_LINE, fp)) {
		if (g_str_has_prefix(buf, COMMENT_PREFIX))
			goto out;

		if (g_str_has_prefix(buf, COMMENT_START_PREFIX)) {
			is_comment = 1;
			goto out;
		}
		if (g_str_has_prefix(buf, COMMENT_END_PREFIX)) {
			is_comment = 0;
			goto out;
		}	
		if (!is_comment) {
			assert(sshfsm.host_num + 1 <= sshfsm.host_num_max);
			if (!sshfsm.host_arr) {
				sshfsm.host_arr = (struct host **) 
				sshfsm_calloc(sshfsm.host_num_max, sizeof(struct host *));
				sshfsm.host_arr_empty_idx = 0;
			}
			buf[strlen(buf)-1] = '\0';
			hostp = sshfsm.host_arr[sshfsm.host_arr_empty_idx] =
				(struct host *) malloc(sizeof(struct host));
			cp = strdup(buf);
			tmp = strtok(cp, delimiters);
			if (!tmp)
				goto out;
			else 
				hostp->host_name = strdup(tmp);
			
			tmp = strtok(NULL, delimiters);
			if (!tmp)
				hostp->base_path = strdup("");
			else {
				if (tmp[strlen(tmp) - 1] != '/') {
					s = (char *) 
						sshfsm_calloc(strlen(tmp) + 2,
									  sizeof(char));
					snprintf(s, strlen(tmp)+2, "%s/", tmp);
					hostp->base_path = s;
				} else
					hostp->base_path = strdup(tmp);
			}
			sshfsm.host_num ++;
			do {
				sshfsm.host_arr_empty_idx = (sshfsm.host_arr_empty_idx + 1)
											% sshfsm.host_num_max;
			} while (sshfsm.host_arr[sshfsm.host_arr_empty_idx] != NULL);
			free(cp);
		}
	  out:
		memset(buf, 0, MAX_LINE);
	}

	fclose(fp);
	free(buf);
	g_free(filename);
	return 0;
}

static char * get_real_path(const char *path)
{
	char *real_path = NULL, *buf = NULL;
	char *s = g_strdup(path);
	struct passwd *pwdp;

	if (s[strlen(s) - 1] == '/')
		s[strlen(s) - 1] = '\0';

	if (s[0] == '/')
		return s;
	
	if (s[0] == '~' && s[1] == '/') {
		pwdp = getpwuid(getuid());
		real_path = g_strdup_printf("%s%s", pwdp->pw_dir, s + 1);
		goto out;
	}

	if (s[0] == '.' && s[1] == '/') {
		buf = sshfsm_calloc(MAX_LINE, sizeof(char));
		buf = getcwd(buf, MAX_LINE);
		if (!buf) {
			real_path = NULL;
			goto out;
		}
		real_path = g_strdup_printf("%s%s", buf, s + 1);
		goto out;
	}

	buf = sshfsm_calloc(MAX_LINE, sizeof(char));
	buf = getcwd(buf, MAX_LINE);
	if (!buf) {
		real_path = NULL;
		goto out;
	}
	real_path = g_strdup_printf("%s/%s", buf, s);

  out:
	free(buf);
	g_free(s);
	return real_path;
}

static char * get_parent_dir(const char *path)
{
	int len, i;

	if (strcmp(path, "/") == 0)
		return g_strdup(path);
	
	len = strlen(path);
	i = path[len - 1] == '/' ? 2 : 1;
	while (path[len - i] != '/' && i < len)
		i++;
	
	if (len == i)
		return g_strdup("/");
	
	return g_strndup(path, len - i);
}

#ifdef SSHFSM_USE_CTRL
static int ctrl_umount()
{
	execlp(FUSERMOUNT, FUSERMOUNT, "-u", ctrl.mountpoint, NULL);
	return 0;
}

static int ctrl_add()
{
	int sockfd, len, res, i, err;
	struct sockaddr_un serv_addr;
	struct stat stbuf;
	char buf[CTRL_BUF_MAX_LEN];
	char *ctrl_file = g_strdup_printf("%s/sshfsm.%u", CTRL_FILE_PATH,
					  			      g_str_hash(ctrl.mountpoint));
	
	if (stat(ctrl_file, &stbuf) == -1) {
		g_free(ctrl_file);
		if (errno == ENOENT) {
			fprintf(stderr, "Control channel for %s not found\n",
					ctrl.mountpoint);
		}
		return -1;
	}

	if (sshfsm.hosts_file != NULL) {
		if (parse_hosts_file(sshfsm.hosts_file) == -1)
			return -1;
	}

	if (sshfsm.host_num == 0) {
		fprintf(stderr, "missing host or hosts_file to add\n");
		return -1;
	}
	sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sockfd == -1) {
		fprintf(stderr, "sshfsm: socket open failed\n");
		return -1;
	}
	serv_addr.sun_family = AF_UNIX;
	assert(strlen(ctrl_file) < 108);
	strncpy(serv_addr.sun_path, ctrl_file, 108);

	len = strlen(serv_addr.sun_path) + sizeof(serv_addr.sun_family);
	if (connect(sockfd, (struct sockaddr *) &serv_addr, len) == -1) {
		fprintf(stderr, "sshfsm: socket connect failed\n");
		return -1;
	}
	
	/* Pack data */
	for (i = 0; i < sshfsm.host_num; i++) {
		memset(buf, 0, CTRL_BUF_MAX_LEN);
		snprintf(buf, CTRL_BUF_MAX_LEN, "%d:%s:%s", CTRL_TYPE_ADD, 
			 sshfsm.host_arr[i]->host_name, sshfsm.host_arr[i]->base_path);
		res = send(sockfd, buf, strlen(buf), 0);
		if (res == -1) {
			fprintf(stderr, "sshfsm: socket send failed\n");
			return -1;
		}

		memset(buf, 0, CTRL_BUF_MAX_LEN);
		res = recv(sockfd, buf, CTRL_BUF_MAX_LEN, 0);
		if (res == -1) {
			fprintf(stderr, "sshfsm: socket recv failed\n");
			return -1;
		}
		err = atoi(buf);
		if (err)
			fprintf(stderr, "Add %s:%s failed: %s\n", 	
			sshfsm.host_arr[i]->host_name, sshfsm.host_arr[i]->base_path,
			strerror(err));
		else
			fprintf(stderr, "Successfully add %s:%s\n",
			sshfsm.host_arr[i]->host_name, sshfsm.host_arr[i]->base_path);
	}

	memset(buf, 0, CTRL_BUF_MAX_LEN);
	snprintf(buf, CTRL_BUF_MAX_LEN, "%d", CTRL_TYPE_FINISH); 
	res = send(sockfd, buf, strlen(buf), 0);
	g_free(ctrl_file);
	close(sockfd);
	return 0;
}

static int ctrl_remove()
{
	int sockfd, len, res, i, err;
	struct sockaddr_un serv_addr;
	struct stat stbuf;
	char buf[CTRL_BUF_MAX_LEN];
	char *ctrl_file = g_strdup_printf("%s/sshfsm.%u", CTRL_FILE_PATH,
							    g_str_hash(ctrl.mountpoint));
	if (stat(ctrl_file, &stbuf) == -1) {
		g_free(ctrl_file);
		if (errno == ENOENT) {
			fprintf(stderr, "Control channel for %s not found\n",
					ctrl.mountpoint);
		}
		return -1;
	}
	if (sshfsm.hosts_file != NULL) {
		if (parse_hosts_file(sshfsm.hosts_file) == -1)
			return -1;
	}
	sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sockfd == -1) {
		fprintf(stderr, "sshfsm: socket open failed\n");
		return -1;
	}
	serv_addr.sun_family = AF_UNIX;
	assert(strlen(ctrl_file) < 108);
	strncpy(serv_addr.sun_path, ctrl_file, 108);

	len = strlen(serv_addr.sun_path) + sizeof(serv_addr.sun_family);
	if (connect(sockfd, (struct sockaddr *) &serv_addr, len) == -1) {
		fprintf(stderr, "sshfsm: socket connect failed\n");
		return -1;
	}
	
	/* Pack data */
	for (i = 0; i < sshfsm.host_num; i++) {
		memset(buf, 0, CTRL_BUF_MAX_LEN);
		snprintf(buf, CTRL_BUF_MAX_LEN, "%d:%s:%s", CTRL_TYPE_REMOVE, 
			 sshfsm.host_arr[i]->host_name, sshfsm.host_arr[i]->base_path);
		res = send(sockfd, buf, strlen(buf), 0);
		if (res == -1) {
			fprintf(stderr, "sshfsm: socket send failed\n");
			return -1;
		}

		memset(buf, 0, CTRL_BUF_MAX_LEN);
		res = recv(sockfd, buf, CTRL_BUF_MAX_LEN, 0);
		if (res == -1) {
			fprintf(stderr, "sshfsm: socket recv failed\n");
			return -1;
		}
		err = atoi(buf);
		if (err)
			fprintf(stderr, "Remove %s:%s failed: %s\n", 	
			sshfsm.host_arr[i]->host_name, sshfsm.host_arr[i]->base_path,
			strerror(err));
		else
			fprintf(stderr, "Successfully remove %s:%s\n",
			sshfsm.host_arr[i]->host_name, sshfsm.host_arr[i]->base_path);
	}

	memset(buf, 0, CTRL_BUF_MAX_LEN);
	snprintf(buf, CTRL_BUF_MAX_LEN, "%d", CTRL_TYPE_FINISH); 
	res = send(sockfd, buf, strlen(buf), 0);
	g_free(ctrl_file);
	close(sockfd);
	return 0;
}

static int ctrl_process_add(int sockfd, const char *hoststr)
{
	int err;
	char buf[4];
	
	err = host_join(hoststr);
	memset(buf, 0, 4);
	snprintf(buf, 4, "%d", err);
	err = send(sockfd, buf, strlen(buf), 0);
	if (err == -1)
		fprintf(stderr, "sshfsm: socket write failed\n");

	return err;
}

static int ctrl_process_remove(int sockfd, const char *hoststr)
{
	int err;
	char buf[4];
	
	err = host_leave(hoststr);
	memset(buf, 0, 4);
	snprintf(buf, 4, "%d", err);
	err = send(sockfd, buf, strlen(buf), 0);
	if (err == -1)
		fprintf(stderr, "sshfsm: socket write failed\n");

	return err;
}

static void * ctrl_process_request(void *data)
{	
	struct sockaddr_un clnt_addr;
	int sockfd, clnt_len, res;
	char buf[CTRL_BUF_MAX_LEN];
	const char delimiters[] = ":";
	char *token;
	int ctrl_type;
	(void) data;

	/* TODO: Error Handling */
	clnt_len = sizeof(clnt_addr);
	do {
		sockfd = accept(sshfsm.ctrl_fd, (struct sockaddr *) &clnt_addr, 
					   (socklen_t *) &clnt_len);
		if (sockfd == -1)
			fprintf(stderr, "sshfsm: socket accept failed\n");
		
		do {	
			memset(buf, 0, CTRL_BUF_MAX_LEN);
			res = recv(sockfd, buf, CTRL_BUF_MAX_LEN, 0);
			if (res == -1)
				fprintf(stderr, "sshfsm: socket read failed\n");
			token = strtok(buf, delimiters);
			ctrl_type = atoi(token);
			switch (ctrl_type) {
				case CTRL_TYPE_ADD:
					res = ctrl_process_add(sockfd, buf+strlen(token)+1);
					break;

				case CTRL_TYPE_REMOVE:
					res = ctrl_process_remove(sockfd, buf+strlen(token)+1);
					break;
				
				case CTRL_TYPE_FINISH:
					goto out;

				default:
					fprintf(stderr, "sshfsm: internal error\n");
					break;
			}
		} while(res >= 0);
	  
	  out:	
		close(sockfd);
	} while (1);
	
	return NULL;
}

static int ctrl_start_processing_thread()
{
	pthread_t thread_id;
	sigset_t oldset, newset;
	int err;
	
	if (sshfsm.ctrl_processing_thread_started)
		return 0;

	sigemptyset(&newset); sigaddset(&newset, SIGTERM);
	sigaddset(&newset, SIGINT);
	sigaddset(&newset, SIGHUP);
	sigaddset(&newset, SIGQUIT);
	pthread_sigmask(SIG_BLOCK, &newset, &oldset);
	err = pthread_create(&thread_id, NULL, ctrl_process_request, NULL);
	if (err) {
		fprintf(stderr, "failed to create thread: %s\n", strerror(err));
		return -1;
	}
	pthread_detach(thread_id);
	pthread_sigmask(SIG_SETMASK, &oldset, NULL);
	sshfsm.ctrl_processing_thread_started = 1;
	return 0;
}

static int ctrl_init()
{	
	struct sockaddr_un serv_addr;
	int len;

	sshfsm.ctrl_file = g_strdup_printf("%s/sshfsm.%u", CTRL_FILE_PATH,
									   g_str_hash(sshfsm.mountpoint));

	sshfsm.ctrl_fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sshfsm.ctrl_fd == -1) {
		fprintf(stderr, "sshfsm: control socket open failed\n\n");
		return -1;
	}

	memset(&serv_addr, 0, sizeof(struct sockaddr_un));
	serv_addr.sun_family = AF_UNIX;

	assert(strlen(sshfsm.ctrl_file) < 108);
	strncpy(serv_addr.sun_path, sshfsm.ctrl_file, 108);
	unlink(serv_addr.sun_path);
	
	len = strlen(serv_addr.sun_path) + sizeof(serv_addr.sun_family);
	if (bind(sshfsm.ctrl_fd, (struct sockaddr *) &serv_addr, len) == -1) {
		fprintf(stderr, "sshfsm: socket bind failed\n");
		return -1;
	}
	if (listen(sshfsm.ctrl_fd, CTRL_MAX_CONN) == -1) {
		fprintf(stderr, "sshfsm: socket listen failed\n");
		return -1;
	}
	
	sshfsm.ctrl_processing_thread_started = 0;
	return ctrl_start_processing_thread();
}

static void ctrl_destroy()
{	
	close(sshfsm.ctrl_fd);
	unlink(sshfsm.ctrl_file);
	g_free(sshfsm.ctrl_file);
}

static int sshfsm_do_ctrl(struct fuse_args *args) 
{	
	int res, i;
	struct host *hostp;

	if (fuse_parse_cmdline(args, &(ctrl.mountpoint), NULL, NULL) == -1)
		return -1;

	if (!ctrl.mountpoint) {
		fprintf(stderr, "sshfsm: missing mountpoint\n");
		return -1;
	}

	sshfsm.conf_dir = sshfsm.conf_dir ? sshfsm.conf_dir : 
					  get_real_path(DEFAULT_CONF_PATH); 

	switch (ctrl.type) {
		case CTRL_TYPE_UMOUNT:
			res = ctrl_umount();
			break;

		case CTRL_TYPE_ADD:
			res = ctrl_add();
			break;

		case CTRL_TYPE_REMOVE:
			res = ctrl_remove();
			break;

		default:
			fprintf(stderr, "sshfsm: unknow control type\n");
			res = -1;
	}
	free(ctrl.mountpoint);
	fuse_opt_free_args(args);
	for (i = 0; i < sshfsm.host_num; i++) {
		hostp = sshfsm.host_arr[i];
		free(hostp->host_name);
		free(hostp->base_path);
	}
	free(sshfsm.host_arr);
	free(sshfsm.conf_dir);
	return res;
}
#endif

#ifdef SSHFSM_USE_INIT
#if FUSE_VERSION >= 26
static void *sshfsm_init(struct fuse_conn_info *conn)
#else
	static void *sshfsm_init(void)
#endif
{	
#if FUSE_VERSION >= 26
	/* Readahead should be done by kernel or sshfsm but not both */
	if (conn->async_read)
		sshfsm.sync_read = 1;
#endif

	if (sshfsm.detect_uid)
		sftp_detect_uid_all();

#ifdef SSHFSM_USE_CTRL
	if (sshfsm.ctrl_on) {
		if (ctrl_init() == -1)
			fprintf(stderr, "sshfsm: statring control failed\n");
	}
#endif

	start_processing_thread_all();
	return NULL;
}
#endif

#ifdef SSHFSM_USE_INIT
static void sshfsm_destroy(void *data)
{	
	int i;
	struct host *hostp;
	(void) data;
	cache_destroy();
	table_destroy();
#ifdef SSHFSM_USE_CTRL
	if (sshfsm.ctrl_on)
		ctrl_destroy();
#endif
	for (i = 0; i < sshfsm.host_num; i++) {
		hostp = sshfsm.host_arr[i];	
		free(hostp->host_name);
		free(hostp->base_path);
		g_hash_table_destroy(hostp->reqtab);
		pthread_mutex_destroy(&hostp->lock);
		pthread_mutex_destroy(&hostp->lock_write);
	}
	free(sshfsm.host_arr);
	g_free(sshfsm.conf_dir);
	g_free(sshfsm.hosts_file);
	g_free(sshfsm.mountpoint);
}
#endif

static int sftp_request_wait(int idx, struct request *req, uint8_t type,
                             uint8_t expect_type, struct buffer *outbuf)
{
	int err;
	struct host *hostp = sshfsm.host_arr[idx];

	if (req->error) {
		err = req->error;
		goto out;
	}
	while (sem_wait(&req->ready));
	if (req->error) {
		err = req->error;
		goto out;
	}
	err = -EIO;
	if (req->reply_type != expect_type &&
	    req->reply_type != SSH_FXP_STATUS) {
		fprintf(stderr, "protocol error\n");
		goto out;
	}
	if (req->reply_type == SSH_FXP_STATUS) {
		uint32_t serr;
		if (buf_get_uint32(&req->reply, &serr) == -1)
			goto out;

		switch (serr) {
		case SSH_FX_OK:
			if (expect_type == SSH_FXP_STATUS)
				err = 0;
			else
				err = -EIO;
			break;

		case SSH_FX_EOF:
			if (type == SSH_FXP_READ || type == SSH_FXP_READDIR)
				err = MY_EOF;
			else
				err = -EIO;
			break;

		default:
			err = -sftp_error_to_errno(serr);
		}
	} else {
		buf_init(outbuf, req->reply.size - req->reply.len);
		buf_get_mem(&req->reply, outbuf->p, outbuf->size);
		err = 0;
	}

out:
	if (req->end_func) {
		pthread_mutex_lock(&hostp->lock);
		req->end_func(req);
		pthread_mutex_unlock(&hostp->lock);
	}
	request_free(req);
	return err;
}

static int sftp_request_send(int idx, uint8_t type, struct iovec *iov, 
							 size_t count, request_func begin_func, 
							 request_func end_func, int want_reply, 
							 void *data, struct request **reqp)
{
	int err;
	uint32_t id;
	struct request *req = g_new0(struct request, 1);
	struct host *hostp = sshfsm.host_arr[idx];

	req->want_reply = want_reply;
	req->end_func = end_func;
	req->data = data;
	sem_init(&req->ready, 0, 0);
	buf_init(&req->reply, 0);
	pthread_mutex_lock(&hostp->lock);
	if (begin_func)
		begin_func(req);
	id = sftp_get_id(idx);
	err = start_processing_thread(idx);
	if (err) {
		pthread_mutex_unlock(&hostp->lock);
		goto out;
	}
	req->len = iov_length(iov, count) + 9;
	sshfsm.outstanding_len += req->len;
	while (sshfsm.outstanding_len > sshfsm.max_outstanding_len)
		pthread_cond_wait(&sshfsm.outstanding_cond, &hostp->lock);

	g_hash_table_insert(hostp->reqtab, GUINT_TO_POINTER(id), req);
	if (sshfsm.debug)
		gettimeofday(&req->start, NULL);
	DEBUG("[%05i] %s\n", id, type_name(type));
	pthread_mutex_unlock(&hostp->lock);

	err = -EIO;
	if (sftp_send_iov(idx, type, id, iov, count) == -1) {
		pthread_mutex_lock(&hostp->lock);
		g_hash_table_remove(hostp->reqtab, GUINT_TO_POINTER(id));
		pthread_mutex_unlock(&hostp->lock);
		goto out;
	}
	if (want_reply)
		*reqp = req;
	return 0;

out:
	req->error = err;
	if (!want_reply)
		sftp_request_wait(idx, req, type, 0, NULL);
	else
		*reqp = req;

	return err;
}


static int sftp_request_iov(int idx, uint8_t type, struct iovec *iov, 
							size_t count, uint8_t expect_type, 
							struct buffer *outbuf)
{
	struct request *req;
	sftp_request_send(idx, type, iov, count, NULL, NULL, expect_type, 
					  NULL, &req);
	if (expect_type == 0)
		return 0;

	return sftp_request_wait(idx, req, type, expect_type, outbuf);
}

static int sftp_request(int idx, uint8_t type, const struct buffer *buf,
                        uint8_t expect_type, struct buffer *outbuf)
{
	struct iovec iov;

	buf_to_iov(buf, &iov);
	return sftp_request_iov(idx, type, &iov, 1, expect_type, outbuf);
}

static int host_getattr(int idx, const char *path, struct stat *stbuf)
{
	int err;
	struct buffer buf, outbuf;

	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	err = sftp_request(idx, sshfsm.follow_symlinks ? 
					   SSH_FXP_STAT : SSH_FXP_LSTAT,
			   		   &buf, SSH_FXP_ATTRS, &outbuf);
	if (!err) {
		if (buf_get_attrs(&outbuf, stbuf, NULL) == -1)
			err = -EIO;
		buf_free(&outbuf);
	} 
	buf_free(&buf);
	return err;
}

struct getattr_thread_data {
	const char *path;
	int idx;
	struct stat stbuf;
	int err;
};

static void * host_getattr_thread(void *data)
{	
	struct getattr_thread_data *datap
		= (struct getattr_thread_data *) data;
	datap->err = host_getattr(datap->idx, datap->path, &(datap->stbuf));
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_getattr(const char *path, struct stat *stbuf)
{
	struct table_entry *entry = NULL;
	struct getattr_thread_data *dat = NULL;
	GSList *idx_lst = NULL;
	pthread_t *threads = NULL;
	pthread_attr_t attr;
	int err, i, retval, i_first = -1, ret = -INT_MAX;
	
	char *parent_dir = get_parent_dir(path);
	entry = table_lookup_r(parent_dir);
	assert(entry != NULL);
	g_free(parent_dir);
	if (!entry)
		return -EIO;
	
	dat = (struct getattr_thread_data *) sshfsm_malloc(
		 	sizeof(struct getattr_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, host_getattr_thread, 
							 &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			ret = -err;
			goto out;
		}
	}
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			ret = err;
			goto out;
		}
		if (retval == 0) {
			i_first = i_first == -1 ? i : i_first;
			if (S_ISDIR(dat[i].stbuf.st_mode))
				idx_lst = g_slist_prepend(idx_lst, entry->idx_arr + i);
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
 	if (ret == 0) {
		*stbuf = dat[i_first].stbuf;
		table_tlb_update(path, dat[i_first].idx);
		table_tlb_clean();
		if (S_ISDIR(stbuf->st_mode)) {
			idx_lst = g_slist_reverse(idx_lst);
			table_update(path, idx_lst);
		}
	}
	g_slist_free(idx_lst);
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_getattr(const char *path, struct stat *stbuf)
{	
	int idx = table_tlb_lookup(path);
	if (idx != -1) {
		if (sshfsm.host_arr[idx] != NULL) {
			if (!host_getattr(idx, path, stbuf)) {
				table_tlb_clean();
				return 0;
			}
		}
		table_tlb_purge(path);
	}
	return batch_getattr(path, stbuf);
}

static int count_components(const char *p)
{
	int ctr;

	for (; *p == '/'; p++);
	for (ctr = 0; *p; ctr++) {
		for (; *p && *p != '/'; p++);
		for (; *p == '/'; p++);
	}
	return ctr;
}

static void strip_common(const char **sp, const char **tp)
{
	const char *s = *sp;
	const char *t = *tp;
	do {
		for (; *s == '/'; s++);
		for (; *t == '/'; t++);
		*tp = t;
		*sp = s;
		for (; *s == *t && *s && *s != '/'; s++, t++);
	} while ((*s == *t && *s) || (!*s && *t == '/') || (*s == '/' && !*t));
}

static void transform_symlink(int idx, const char *path, char **linkp)
{
	const char *l = *linkp;
	const char *b = sshfsm.host_arr[idx]->base_path;
	char *newlink;
	char *s;
	int dotdots;
	int i;

	if (l[0] != '/' || b[0] != '/')
		return;

	strip_common(&l, &b);
	if (*b)
		return;

	strip_common(&l, &path);
	dotdots = count_components(path);
	if (!dotdots)
		return;
	dotdots--;

	newlink = sshfsm_malloc(dotdots * 3 + strlen(l) + 2);
	for (s = newlink, i = 0; i < dotdots; i++, s += 3)
		strcpy(s, "../");

	if (l[0])
		strcpy(s, l);
	else if (!dotdots)
		strcpy(s, ".");
	else
		s[0] = '\0';

	free(*linkp);
	*linkp = newlink;
}

static int host_readlink(int idx, const char *path, 
						 char *linkbuf, size_t size)
{
	int err;
	struct buffer buf, name;

	assert(size > 0);
	if (sshfsm.host_arr[idx]->server_version < 3)
		return -EPERM;
	
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	err = sftp_request(idx, SSH_FXP_READLINK, &buf, SSH_FXP_NAME, &name);
	if (!err) {
		uint32_t count;
		char *link;
		err = -EIO;
		if(buf_get_uint32(&name, &count) != -1 && count == 1 
		   && buf_get_string(&name, &link) != -1) {
			if (sshfsm.transform_symlinks)
				transform_symlink(idx, path, &link);
			strncpy(linkbuf, link, size - 1);
			linkbuf[size - 1] = '\0';
			free(link);
			err = 0;
		}
		buf_free(&name);
	}
	buf_free(&buf);

	return err;
}

struct readlink_thread_data {
	const char *path;
	int idx;
	char *linkbuf;
	size_t size;
	int err;
};

static void * host_readlink_thread(void *data)
{
	struct readlink_thread_data *datap
		= (struct readlink_thread_data *) data;
	datap->err = host_readlink(datap->idx, datap->path, 
			     			   datap->linkbuf, datap->size);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_readlink(const char *path, char *linkbuf, size_t size)
{
	struct table_entry *entry;
	struct readlink_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(path);
	if (!entry)
		return -EIO;
	
	dat = (struct readlink_thread_data *) sshfsm_malloc(
					 sizeof(struct readlink_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].idx = entry->idx_arr[i];
		dat[i].linkbuf = linkbuf;
		dat[i].size = size;
		err = pthread_create(&threads[i], &attr, 
							 host_readlink_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			ret = -err;
			goto out;
		}
	}
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			ret = -err;
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
		if (ret == 0)
			goto out;
	}
 
 out:
 	if (ret == 0)
		table_tlb_update(path, entry->idx_arr[i]);
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_readlink(const char *path, char *linkbuf, size_t size)
{
	int idx = table_tlb_lookup(path);
	if (idx != -1) {
		assert(sshfsm.host_arr[idx] != NULL);
		if (!host_readlink(idx, path, linkbuf, size)) {
			table_tlb_clean();
			return 0;
		} else
			table_tlb_purge(path);
	}
	return batch_readlink(path, linkbuf, size);
}

static int host_getdir(int idx, const char *path, GSList **lst,
					   dir_entry_fill_t filler)
{
	int err;
	struct buffer buf, handle;

	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	err = sftp_request(idx, SSH_FXP_OPENDIR, &buf, SSH_FXP_HANDLE, &handle);
	if (!err) {
		int err2;
		buf_finish(&handle);
		do {
			struct buffer name;
			err = sftp_request(idx, SSH_FXP_READDIR, &handle, 
							   SSH_FXP_NAME, &name);
			if (!err) {
				if (buf_get_entries(&name, lst, filler) == -1)
					err = -EIO;
				buf_free(&name);
			}
		} while (!err);
		if (err == MY_EOF)
			err = 0;

		err2 = sftp_request(idx, SSH_FXP_CLOSE, &handle, 0, NULL);
		if (!err)
			err = err2;
		buf_free(&handle);
	}
	buf_free(&buf);
	return err;
}

struct getdir_thread_data {
	const char *path;
	int idx;
	GSList *lst;
	int err;
};

static void * host_getdir_thread(void *data)
{	
	struct getdir_thread_data *datap
		= (struct getdir_thread_data *) data;
	datap->err = host_getdir(datap->idx, datap->path, &datap->lst,
							 dir_entry_fill);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

struct dir_entry_list_handle {
	fuse_cache_dirh_t h;
	fuse_cache_dirfil_t filler;
	GHashTable *filter;
};

static void dir_entry_list_handler(void *entry, void *handle)
{
	struct dir_entry *ep = (struct dir_entry *) entry;
	struct dir_entry_list_handle *hp = 
		(struct dir_entry_list_handle *) handle;
	if (g_hash_table_lookup(hp->filter, ep->name) == NULL) {
		char *data = g_strdup("");
		g_hash_table_insert(hp->filter, g_strdup(ep->name), data);
		hp->filler(hp->h, ep->name, &ep->stbuf);
	}
	g_free(ep->name);
	g_free(ep);
}

static int sshfsm_getdir(const char *path, fuse_cache_dirh_t h,
                         fuse_cache_dirfil_t filler)
{
	struct table_entry *entry = NULL;
	struct getdir_thread_data *dat = NULL;
	struct dir_entry_list_handle handle;
	GHashTable *filter;
	pthread_t *threads = NULL;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;

	filter = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	if (!filter) {
		fprintf(stderr, "failed to create directory entry filter\n");
		return -EIO;
	}
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
  
	dat = (struct getdir_thread_data *) sshfsm_malloc(
		 	sizeof(struct getdir_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].idx = entry->idx_arr[i];
		dat[i].lst = NULL;
		err = pthread_create(&threads[i], &attr, host_getdir_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			ret = -err;
			goto out;
		}
	}
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			ret = -err;
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
	handle.h = h;
	handle.filler = filler;
	handle.filter = filter;
	for (i = 0; i < entry->size; i++) {
		g_slist_foreach(dat[i].lst, dir_entry_list_handler, &handle);
		g_slist_free(dat[i].lst);
	}
  out:
	g_hash_table_destroy(filter);
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int host_mkdir(int idx, const char *path, mode_t mode)
{
	int err;
	struct buffer buf;
	
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_PERMISSIONS);
	buf_add_uint32(&buf, mode);
	err = sftp_request(idx, SSH_FXP_MKDIR, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}


static int sshfsm_mkdir(const char *path, mode_t mode)
{
	struct table_entry *entry;
	GSList *idx_lst = NULL;
	int idx, err, i, ret = -INT_MAX;
	char *parent_dir;
	
	parent_dir = get_parent_dir(path);
	idx = table_tlb_lookup(parent_dir);
	if (idx != -1) {
		err = host_mkdir(idx, path, mode);
		ret = err > ret ? err : ret;
		if (!err) {
			g_free(parent_dir);
			goto out;
		}
	}

	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	for (i = 0; i < entry->size; i++) {
		if ((idx = entry->idx_arr[i]) == -1)
			continue;
		/* TODO: Load Balancing by mkdir */	
		err = host_mkdir(idx, path, mode);
		ret = err > ret ? err : ret;
		if (!err)
			goto out;	
	}

  out:
 	if (ret == 0) {
		table_tlb_update(path, idx);
		if (S_ISDIR(mode)) {
			idx_lst = g_slist_prepend(idx_lst, &idx);
			table_update(path, idx_lst);
		}
	}
	return ret;
}

static int host_mknod(int idx, const char *path, mode_t mode, dev_t rdev)
{
	int err, err2;
	struct buffer buf, handle;
	(void) rdev;
	
	if ((mode & S_IFMT) != S_IFREG)
		return -EPERM;

	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, SSH_FXF_WRITE | SSH_FXF_CREAT | 
				   SSH_FXF_EXCL);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_PERMISSIONS);
	buf_add_uint32(&buf, mode);
	err = sftp_request(idx, SSH_FXP_OPEN, &buf, 
					   SSH_FXP_HANDLE, &handle);
	if (!err) {
		buf_finish(&handle);
		err2 = sftp_request(idx, SSH_FXP_CLOSE, &handle, 
							SSH_FXP_STATUS, NULL);
		if (!err)
			err = err2;
		buf_free(&handle);
	}
	buf_free(&buf);
	return err;
}

static int sshfsm_mknod(const char *path, mode_t mode, dev_t rdev)
{
	struct table_entry *entry;
	GSList *idx_lst = NULL;	
	char *parent_dir = NULL;
	int idx, err, i, ret = -INT_MAX;
	
	parent_dir = get_parent_dir(path);
	idx = table_tlb_lookup(parent_dir);
	if (idx != -1) {
		err = host_mknod(idx, path, mode, rdev);
		ret = err > ret ? err : ret;
		if (!err) {
			g_free(parent_dir);
			goto out;
		}
	}
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	for (i = 0; i < entry->size; i++) {
		if ((idx = entry->idx_arr[i]) == -1)
			continue;
		
		err = host_mknod(idx, path, mode, rdev);
		ret = err > ret ? err : ret;
		if (!err)
			break;	
	}
	
 out:
	if (ret == 0) {
		table_tlb_update(path, idx);
		if (S_ISDIR(mode)) {
			idx_lst = g_slist_prepend(idx_lst, &idx);
			table_update(path, idx_lst);
		}
	}
	return ret;
}

static int host_symlink(int idx, const char *from, const char *to)
{
	int err;
	struct buffer buf;

	if (sshfsm.host_arr[idx]->server_version < 3)
		return -EPERM;

	/* openssh sftp server doesn't follow standard: link 
	 * target and link name are mixed up, so we must also 
	 * be non-standard :( */
	buf_init(&buf, 0);
	buf_add_string(&buf, from);
	buf_add_path(&buf, idx, to);
	err = sftp_request(idx, SSH_FXP_SYMLINK, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

static int sshfsm_symlink(const char *from, const char *to)
{
	struct table_entry *entry;
	int idx, err, i, ret = -INT_MAX;
	
	idx = table_tlb_lookup(from);
	if (idx != -1) {
		err = host_symlink(idx, from, to);
		ret = err > ret ? err : ret;
		if (!err)
			goto out;
	}

	entry = table_lookup_r(from);
	assert(entry != NULL);
	if (!entry)
		return -EIO;

	for (i = 0; i < entry->size; i++) {
		if ((idx = entry->idx_arr[i]) == -1)
			continue;

		err = host_symlink(idx, from, to);
		ret = err > ret ? err : ret;
		if (!err)
			goto out;	
	}
  
  out:
  	if (ret == 0) {
  		table_tlb_update(from, idx);
		table_tlb_update(to, idx);
	}
	/* TODO: table_update(path, idx_arr); 
	 * Can readlink to instead ?? */
	return ret;
}

static int host_unlink(int idx, const char *path)
{
	int err;
	struct buffer buf;
	
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	err = sftp_request(idx, SSH_FXP_REMOVE, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

struct unlink_thread_data {
	const char *path;
	int idx;
	int err;
};

static void * host_unlink_thread(void *data)
{
	struct unlink_thread_data *datap
		= (struct unlink_thread_data *) data;
	datap->err = host_unlink(datap->idx, datap->path);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_unlink(const char *path)
{
	struct table_entry *entry = NULL;
	struct unlink_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	dat = (struct unlink_thread_data *) sshfsm_malloc(
						 sizeof(struct unlink_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_unlink_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 out:
 	if (ret == 0) {
		table_tlb_purge(path);
		table_remove(path);
	}
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_unlink(const char *path)
{
	return batch_unlink(path);
}

static int host_rmdir(int idx, const char *path)
{
	int err;
	struct buffer buf;
	
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	err = sftp_request(idx, SSH_FXP_RMDIR, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

struct rmdir_thread_data {
	const char *path;
	int idx;
	int err;
};

static void * host_rmdir_thread(void *data)
{
	struct rmdir_thread_data *datap =
		(struct rmdir_thread_data *) data;
	datap->err = host_rmdir(datap->idx, datap->path);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_rmdir(const char *path)
{
	struct table_entry *entry = NULL;
	struct rmdir_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;

	dat = (struct rmdir_thread_data *) sshfsm_malloc(
						 sizeof(struct rmdir_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_rmdir_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
 	if (ret == 0) {
		table_tlb_purge(path);
		table_remove(path);
	}
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_rmdir(const char *path)
{
	return batch_rmdir(path);
}

static int host_do_rename(int idx, const char *from, const char *to)
{
	int err;
	struct buffer buf;

	buf_init(&buf, 0);
	buf_add_path(&buf, idx, from);
	buf_add_path(&buf, idx, to);
	err = sftp_request(idx, SSH_FXP_RENAME, &buf, SSH_FXP_STATUS, NULL);
	buf_free(&buf);
	return err;
}

static void random_string(char *str, int length)
{
	int i;
	for (i = 0; i < length; i++)
		*str++ = (char)('0' + rand_r(&sshfsm.randseed) % 10);
	*str = '\0';
}

static int host_rename(int idx, const char *from, const char *to)
{
	int err;
	err = host_do_rename(idx, from, to);
	if (err == -EPERM && sshfsm.rename_workaround) {
		size_t tolen = strlen(to);
		if (tolen + RENAME_TEMP_CHARS < PATH_MAX) {
			int tmperr;
			char totmp[PATH_MAX];
			strcpy(totmp, to);
			random_string(totmp + tolen, RENAME_TEMP_CHARS);
			tmperr = host_do_rename(idx, to, totmp);
			if (!tmperr) {
				err = host_do_rename(idx, from, to);
				if (!err)
					err = host_unlink(idx, totmp);
				else
					host_do_rename(idx, totmp, to);
			}
		}
	}
	return err;
}

struct rename_thread_data {
	const char *from;
	const char *to;
	int idx;
	int err;
};

static void * host_rename_thread(void *data)
{
	struct rename_thread_data *datap
		= (struct rename_thread_data *) data;
	datap->err = host_rename(datap->idx, datap->from, datap->to);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_rename(const char *from, const char *to)
{
	struct table_entry *entry = NULL;
	struct rename_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(from);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	dat = (struct rename_thread_data *) sshfsm_malloc(
						 sizeof(struct rename_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].from = from;
		dat[i].to = to;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_rename_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
 	if (ret == 0) {
		i = 0;
		while (dat[i].err != 0 && i < entry->size)
			i++;
		table_tlb_update(to, entry->idx_arr[i]);
		table_tlb_purge(from);
		table_remove(from);
	}
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_rename(const char *from, const char *to)
{
	return batch_rename(from, to);
}

static int host_chmod(int idx, const char *path, mode_t mode)
{	
	int err;
	struct buffer buf;
	
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_PERMISSIONS);
	buf_add_uint32(&buf, mode);
	err = sftp_request(idx, SSH_FXP_SETSTAT, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

struct chmod_thread_data {
	const char *path;
	mode_t *mode;
	int idx;
	int err;
};

static void * host_chmod_thread(void *data)
{	
	struct chmod_thread_data *datap
		= (struct chmod_thread_data *) data;
	datap->err = host_chmod(datap->idx, datap->path, *(datap->mode));
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_chmod(const char *path, mode_t mode)
{
	struct table_entry *entry = NULL;
	struct chmod_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	dat = (struct chmod_thread_data *) sshfsm_malloc(
						 sizeof(struct chmod_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].mode = &mode;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_chmod_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_chmod(const char *path, mode_t mode)
{
	int idx, err = 0;
	struct stat stbuf;
	
	idx = table_tlb_lookup(path);
	if (idx != -1) {
		err = host_getattr(idx, path, &stbuf);
		if (!err && !S_ISDIR(stbuf.st_mode)) {
			err = host_chmod(idx, path, mode);
			return err;
		} else if (err)
			table_tlb_purge(path);
	}
	return batch_chmod(path, mode);
}

static int host_chown(int idx, const char *path, uid_t uid, gid_t gid)
{
	int err;
	struct buffer buf;
	
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_UIDGID);
	buf_add_uint32(&buf, uid);
	buf_add_uint32(&buf, gid);
	err = sftp_request(idx, SSH_FXP_SETSTAT, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

struct chown_thread_data {
	const char *path;
	uid_t uid;
	gid_t gid;
	int idx;
	int err;
};

static void * host_chown_thread(void *data)
{	
	struct chown_thread_data *datap
		= (struct chown_thread_data *) data;
	datap->err = host_chown(datap->idx, datap->path, datap->uid, datap->gid);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_chown(const char *path, uid_t uid, gid_t gid)
{
	struct table_entry *entry = NULL;
	struct chown_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	dat = (struct chown_thread_data *) sshfsm_malloc(
						 sizeof(struct chown_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].uid = uid;
		dat[i].gid = gid;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_chown_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_chown(const char *path, uid_t uid, gid_t gid)
{
	int idx, err = 0;
	struct stat stbuf;
	
	idx = table_tlb_lookup(path);
	if (idx != -1) {
		err = host_getattr(idx, path, &stbuf);
		if (!err && !S_ISDIR(stbuf.st_mode)) {
			err = host_chown(idx, path, uid, gid);
			return err;
		} else if (err)
			table_tlb_purge(path);
	}
	return batch_chown(path, uid, gid);
}

static int host_truncate_workaround(int idx, const char *path,
					off_t size, struct fuse_file_info *fi);

static int host_truncate(int idx, const char *path, off_t size)
{
	int err;
	struct buffer buf;
	
	sshfsm.host_arr[idx]->modifver ++;
	if (size == 0 || sshfsm.truncate_workaround)
		return host_truncate_workaround(idx, path, size, NULL);

	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_SIZE);
	buf_add_uint64(&buf, size);
	err = sftp_request(idx, SSH_FXP_SETSTAT, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);
	return err;
}

struct truncate_thread_data {
	const char *path;
	off_t size;
	int idx;
	int err;
};

static void * host_truncate_thread(void *data)
{
	struct truncate_thread_data *datap
		= (struct truncate_thread_data *) data;
	datap->err = host_truncate(datap->idx, datap->path, datap->size);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_truncate(const char *path, off_t size)
{
	struct table_entry *entry = NULL;
	struct truncate_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;

	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	dat = (struct truncate_thread_data *) sshfsm_malloc(
						 sizeof(struct truncate_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].size = size;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_truncate_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_truncate(const char *path, off_t size)
{
	int idx, err = 0;
	
	idx = table_tlb_lookup(path);
	if (idx != -1) {
		err = host_truncate(idx, path, size);
		if (!err)
			return err;
		else
			table_tlb_purge(path);
	}
	return batch_truncate(path, size);
}

static int host_utime(int idx, const char *path, struct utimbuf *ubuf)
{
	int err;
	struct buffer buf;

	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_ACMODTIME);
	buf_add_uint32(&buf, ubuf->actime);
	buf_add_uint32(&buf, ubuf->modtime);
	err = sftp_request(idx, SSH_FXP_SETSTAT, &buf, 
					   SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

struct utime_thread_data {
	const char *path;
	struct utimbuf *ubuf;
	int idx;
	int err;
};

static void * host_utime_thread(void *data)
{
	struct utime_thread_data *datap
		= (struct utime_thread_data *) data;
	datap->err = host_utime(datap->idx, datap->path, datap->ubuf);
	datap->err ? pthread_exit((void *) -1) : pthread_exit((void *) 0);
}

static int batch_utime(const char *path, struct utimbuf *ubuf)
{
	struct table_entry *entry = NULL;
	struct utime_thread_data *dat = NULL;
	pthread_t *threads;
	pthread_attr_t attr;
	int err, i, retval, ret = -INT_MAX;
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	dat = (struct utime_thread_data *) sshfsm_malloc(
						 sizeof(struct utime_thread_data) * entry->size);
	threads = (pthread_t *) sshfsm_malloc(sizeof(pthread_t) * entry->size);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	for (i = 0; i < entry->size; i++) {
		dat[i].path = path;
		dat[i].ubuf = ubuf;
		dat[i].idx = entry->idx_arr[i];
		err = pthread_create(&threads[i], &attr, 
							 host_utime_thread, &dat[i]);
		if (err) {
			fprintf(stderr, "sshfsm: create thread failed: %s\n", 
					strerror(err));
			goto out;
		}
	}
	
	for (i = 0; i < entry->size; i++) {
		err = pthread_join(threads[i], (void *) &retval);
		if (err) {
			fprintf(stderr, "sshfsm: join thread failed: %s\n", 
					strerror(err));
			goto out;
		}
		ret = dat[i].err > ret ? dat[i].err : ret;
	}
 
 out:
	pthread_attr_destroy(&attr);
	free(dat);
	free(threads);
	return ret;
}

static int sshfsm_utime(const char *path, struct utimbuf *ubuf)
{
	int idx, err = 0;
	struct stat stbuf;
	
	idx = table_tlb_lookup(path);
	if (idx != -1) {
		err = host_getattr(idx, path, &stbuf);
		if (!err && !S_ISDIR(stbuf.st_mode)) {
			err = host_utime(idx, path, ubuf);
			return err;
		} else if (err)
			table_tlb_purge(path);
	}
	return batch_utime(path, ubuf);
}

static inline int host_file_is_conn(int idx, struct sshfsm_file *sf)
{
	return sf->connver == sshfsm.host_arr[idx]->connver;
}

static int host_open_common(int idx, const char *path, 
					mode_t mode, struct fuse_file_info *fi)
{
	int err, err2;
	struct buffer buf, outbuf;
	struct stat stbuf;
	struct sshfsm_file *sf;
	struct request *open_req;
	uint32_t pflags = 0;
	struct iovec iov;
	uint8_t type;
	
	if ((fi->flags & O_ACCMODE) == O_RDONLY)
		pflags = SSH_FXF_READ;
	else if((fi->flags & O_ACCMODE) == O_WRONLY)
		pflags = SSH_FXF_WRITE;
	else if ((fi->flags & O_ACCMODE) == O_RDWR)
		pflags = SSH_FXF_READ | SSH_FXF_WRITE;
	else
		return -EINVAL;

	if (fi->flags & O_CREAT)
		pflags |= SSH_FXF_CREAT;

	if (fi->flags & O_EXCL)
		pflags |= SSH_FXF_EXCL;

	if (fi->flags & O_TRUNC)
		pflags |= SSH_FXF_TRUNC;

	sf = g_new0(struct sshfsm_file, 1);
	list_init(&sf->write_reqs);
	pthread_cond_init(&sf->write_finished, NULL);
	/* Assume random read after open */
	sf->is_seq = 0;
	sf->refs = 1;
	sf->next_pos = 0;
	sf->host_idx = idx;
	sf->modifver= sshfsm.host_arr[idx]->modifver;
	sf->connver = sshfsm.host_arr[idx]->connver;
	buf_init(&buf, 0);
	buf_add_path(&buf, idx, path);
	buf_add_uint32(&buf, pflags);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_PERMISSIONS);
	buf_add_uint32(&buf, mode);
	buf_to_iov(&buf, &iov);
	sftp_request_send(idx, SSH_FXP_OPEN, &iov, 1, NULL, 
					  NULL, 1, NULL, &open_req);
	buf_clear(&buf);
	buf_add_path(&buf, idx, path);
	type = sshfsm.follow_symlinks ? SSH_FXP_STAT : 
									SSH_FXP_LSTAT;
	err2 = sftp_request(idx, type, &buf, SSH_FXP_ATTRS, 
						&outbuf);
	if (!err2) {
		if (buf_get_attrs(&outbuf, &stbuf, NULL) == -1)
			err2 = -EIO;
		buf_free(&outbuf);
	}
	err = sftp_request_wait(idx, open_req, SSH_FXP_OPEN, SSH_FXP_HANDLE, 
							&sf->handle);
	if (!err && err2) {
		buf_finish(&sf->handle);
		sftp_request(idx, SSH_FXP_CLOSE, &sf->handle, 0, NULL);
		buf_free(&sf->handle);
		err = err2;
	}

	if (!err) {
		cache_add_attr(path, &stbuf);
		buf_finish(&sf->handle);
		fi->fh = (unsigned long) sf;
	} else {
		cache_invalidate(path);
		g_free(sf);
	}
	buf_free(&buf);
	return err;
}

static int sshfsm_open(const char *path, 
					   struct fuse_file_info *fi)
{	
	struct table_entry *entry = NULL;
	char *parent_dir;
	int idx, i, err = 0, ret = -INT_MAX;
	
	parent_dir = get_parent_dir(path);
	idx = table_tlb_lookup(parent_dir);
	if (idx != -1) {
		err = host_open_common(idx, path, 0, fi);
		if (!err) {
			g_free(parent_dir);
			return err;
		} else
			table_tlb_purge(path);
	}
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	for (i = 0; i < entry->size; i++) {
		if ((idx = entry->idx_arr[i]) == -1)
			continue;

		err = host_open_common(idx, path, 0, fi);
		ret = err > ret ? err : ret;
		if (!err) 
			goto out;
	}
  out:
	if (ret == 0) {
		table_tlb_update(path, idx);
	}
	return ret;
}

static inline struct sshfsm_file *get_sshfsm_file(struct fuse_file_info *fi)
{
	return (struct sshfsm_file *) (uintptr_t) fi->fh;
}

static int sshfsm_flush(const char *path, 
					  struct fuse_file_info *fi)
{
	int err;
	struct sshfsm_file *sf = get_sshfsm_file(fi);
	int idx = sf->host_idx;
	struct host *hostp = sshfsm.host_arr[idx];
	struct list_head write_reqs;
	struct list_head *curr_list;
	(void) path;
	
	if (!host_file_is_conn(idx, sf))
		return -EIO;

	if (sshfsm.sync_write)
		return 0;

	pthread_mutex_lock(&hostp->lock);
	if (!list_empty(&sf->write_reqs)) {
		curr_list = sf->write_reqs.prev;
		list_del(&sf->write_reqs);
		list_init(&sf->write_reqs);
		list_add(&write_reqs, curr_list);
		while (!list_empty(&write_reqs))
			pthread_cond_wait(&sf->write_finished, &hostp->lock);
	}
	err = sf->write_error;
	sf->write_error = 0;
	pthread_mutex_unlock(&hostp->lock);
	return err;
}

static int sshfsm_fsync(const char *path, int isdatasync,
                       struct fuse_file_info *fi)
{
	(void) isdatasync;
	return sshfsm_flush(path, fi);
}

static void sshfsm_file_put(struct sshfsm_file *sf)
{
	sf->refs--;
	if (!sf->refs)
		g_free(sf);
}

static void sshfsm_file_get(struct sshfsm_file *sf)
{
	sf->refs++;
}

static int sshfsm_release(const char *path, 
						  struct fuse_file_info *fi)
{
	struct sshfsm_file *sf = get_sshfsm_file(fi);
	struct buffer *handle = &sf->handle;
	int idx = sf->host_idx;
	if (host_file_is_conn(idx, sf)) {
		sshfsm_flush(path, fi);
		sftp_request(idx, SSH_FXP_CLOSE, handle, 
					 0, NULL);
	}
	buf_free(handle);
	chunk_put_locked(idx, sf->readahead);
	sshfsm_file_put(sf);
	return 0;
}

static int sshfsm_sync_read(struct sshfsm_file *sf, char *rbuf, size_t size, off_t offset)
{
	int err;
	struct buffer buf;
	struct buffer data;
	struct buffer *handle = &sf->handle;
	int idx = sf->host_idx;

	buf_init(&buf, 0);
	buf_add_buf(&buf, handle);
	buf_add_uint64(&buf, offset);
	buf_add_uint32(&buf, size);
	err = sftp_request(idx, SSH_FXP_READ, &buf, 
					   SSH_FXP_DATA, &data);
	if (!err) {
		uint32_t retsize;
		err = -EIO;
		if (buf_get_uint32(&data, &retsize) != -1) {
			if (retsize > size)
				fprintf(stderr, "long read\n");
			else {
				buf_get_mem(&data, rbuf, retsize);
				err = retsize;
			}
		}
		buf_free(&data);
	} else if (err == MY_EOF)
		err = 0;
	buf_free(&buf);
	return err;
}

static void sshfsm_read_end(struct request *req)
{
	struct read_chunk *chunk = (struct read_chunk *) req->data;
	if (req->error)
		chunk->res = req->error;
	else if (req->replied) {
		chunk->res = -EIO;

		if (req->reply_type == SSH_FXP_STATUS) {
			uint32_t serr;
			if (buf_get_uint32(&req->reply, &serr) != -1) {
				if (serr == SSH_FX_EOF)
					chunk->res = 0;
			}
		} else if (req->reply_type == SSH_FXP_DATA) {
			uint32_t retsize;
			if (buf_get_uint32(&req->reply, &retsize) != -1) {
				if (retsize > chunk->size)
					fprintf(stderr, "long read\n");
				else {
					chunk->res = retsize;
					chunk->data = req->reply;
					buf_init(&req->reply, 0);
				}
			}
		} else
			fprintf(stderr, "protocol error\n");
	} else
		chunk->res = -EIO;

	sem_post(&chunk->ready);
	chunk_put(chunk);
}

static void sshfsm_read_begin(struct request *req)
{
	struct read_chunk *chunk = (struct read_chunk *) req->data;
	chunk->refs++;
}

static void sshfsm_send_async_read(struct sshfsm_file *sf,
                                  struct read_chunk *chunk)
{	
	struct buffer buf;
	struct buffer *handle = &sf->handle;
	struct iovec iov;
	int idx = sf->host_idx;
	
	buf_init(&buf, 0);
	buf_add_buf(&buf, handle);
	buf_add_uint64(&buf, chunk->offset);
	buf_add_uint32(&buf, chunk->size);
	buf_to_iov(&buf, &iov);
	sftp_request_send(idx, SSH_FXP_READ, &iov, 1, 
					  sshfsm_read_begin, sshfsm_read_end, 
					  0, chunk, NULL);
	buf_free(&buf);
}

static void submit_read(struct sshfsm_file *sf, size_t size, 
					off_t offset, struct read_chunk **chunkp)
{
	struct read_chunk *chunk = g_new0(struct read_chunk, 1);
	struct host *hostp = sshfsm.host_arr[sf->host_idx];	
	sem_init(&chunk->ready, 0, 0);
	buf_init(&chunk->data, 0);
	chunk->offset = offset;
	chunk->size = size;
	chunk->refs = 1;
	chunk->modifver = sshfsm.host_arr[sf->host_idx]->modifver;
	sshfsm_send_async_read(sf, chunk);
	pthread_mutex_lock(&hostp->lock);
	chunk_put(*chunkp);
	*chunkp = chunk;
	pthread_mutex_unlock(&hostp->lock);
}

static int wait_chunk(int idx, struct read_chunk *chunk, char *buf, 
					  size_t size)
{
	int res;
	while (sem_wait(&chunk->ready));
	res = chunk->res;
	if (res > 0) {
		if ((size_t) res > size)
			res = size;
		buf_get_mem(&chunk->data, buf, res);
		chunk->offset += res;
		chunk->size -= res;
		chunk->res -= res;
	}
	sem_post(&chunk->ready);
	chunk_put_locked(idx, chunk);
	return res;
}

static struct read_chunk *search_read_chunk(struct sshfsm_file *sf, 
											off_t offset)
{
	struct read_chunk *ch = sf->readahead;
	if (ch && ch->offset == offset && 
		ch->modifver == sshfsm.host_arr[sf->host_idx]->modifver) {
		ch->refs++;
		return ch;
	} else
		return NULL;
}

static int sshfsm_async_read(struct sshfsm_file *sf, char *rbuf, 
							 size_t size, off_t offset)
{
	int res = 0;
	size_t total = 0;
	struct read_chunk *chunk;
	struct read_chunk *chunk_prev = NULL;
	size_t origsize = size;
	int curr_is_seq;
	int idx = sf->host_idx;
	struct host *hostp = sshfsm.host_arr[idx];

	pthread_mutex_lock(&hostp->lock);
	curr_is_seq = sf->is_seq;
	sf->is_seq = (sf->next_pos == offset && 
				  sf->modifver == sshfsm.host_arr[sf->host_idx]->modifver);
	sf->next_pos = offset + size;
	sf->modifver = sshfsm.host_arr[sf->host_idx]->modifver;
	chunk = search_read_chunk(sf, offset);
	pthread_mutex_unlock(&hostp->lock);

	if (chunk && chunk->size < size) {
		chunk_prev = chunk;
		size -= chunk->size;
		offset += chunk->size;
		chunk = NULL;
	}

	if (!chunk)
		submit_read(sf, size, offset, &chunk);

	if (curr_is_seq && chunk && chunk->size <= size)
		submit_read(sf, origsize, offset + size, 
					&sf->readahead);

	if (chunk_prev) {
		size_t prev_size = chunk_prev->size;
		res = wait_chunk(idx, chunk_prev, rbuf, prev_size);
		if (res < (int) prev_size) {
			chunk_put_locked(idx, chunk);
			return res;
		}
		rbuf += res;
		total += res;
	}
	res = wait_chunk(idx, chunk, rbuf, size);
	if (res > 0)
		total += res;
	if (res < 0)
		return res;

	return total;
}

static int sshfsm_read(const char *path, char *rbuf, 
	size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct sshfsm_file *sf = get_sshfsm_file(fi);
	int idx = sf->host_idx;
	(void) path;
	
	if (!host_file_is_conn(idx, sf))
		return -EIO;

	if (sshfsm.sync_read)
		return sshfsm_sync_read(sf, rbuf, size, offset);
	else
		return sshfsm_async_read(sf, rbuf, size, offset);
}

static void sshfsm_write_begin(struct request *req)
{
	struct sshfsm_file *sf = (struct sshfsm_file *) req->data;

	sshfsm_file_get(sf);
	list_add(&req->list, &sf->write_reqs);
}

static void sshfsm_write_end(struct request *req)
{
	uint32_t serr;
	struct sshfsm_file *sf = (struct sshfsm_file *) req->data;

	if (req->error)
		sf->write_error = req->error;
	else if (req->replied) {
		if (req->reply_type != SSH_FXP_STATUS) {
			fprintf(stderr, "protocol error\n");
		} else if (buf_get_uint32(&req->reply, &serr) != -1 &&
			 serr != SSH_FX_OK) {
			sf->write_error = -EIO;
		}
	}
	list_del(&req->list);
	pthread_cond_broadcast(&sf->write_finished);
	sshfsm_file_put(sf);
}

static int sshfsm_write(const char *path, const char *wbuf, 
		size_t size, off_t offset, struct fuse_file_info *fi)
{
	int err;
	struct buffer buf;
	struct sshfsm_file *sf = get_sshfsm_file(fi);
	struct buffer *handle = &sf->handle;
	struct iovec iov[2];
	int idx = sf->host_idx;
	(void) path;

	if (!host_file_is_conn(idx, sf))
		return -EIO;
	
	sshfsm.host_arr[idx]->modifver ++;
	buf_init(&buf, 0);
	buf_add_buf(&buf, handle);
	buf_add_uint64(&buf, offset);
	buf_add_uint32(&buf, size);
	buf_to_iov(&buf, &iov[0]);
	iov[1].iov_base = (void *) wbuf;
	iov[1].iov_len = size;
	if (!sshfsm.sync_write && !sf->write_error) {
		err = sftp_request_send(idx, SSH_FXP_WRITE, iov, 2,
					sshfsm_write_begin, sshfsm_write_end,
					0, sf, NULL);
	} else {
		err = sftp_request_iov(idx, SSH_FXP_WRITE, iov, 2, 
							   SSH_FXP_STATUS, NULL);
	}
	buf_free(&buf);
	return err ? err : (int) size;
}

#if FUSE_VERSION >= 25
static int sshfsm_statfs(const char *path, struct statvfs *buf)
{
	(void) path;

	buf->f_namemax = 255;
	buf->f_bsize = sshfsm.blksize;
	/*
	 * df seems to use f_bsize instead of f_frsize, 
	 * so make them the same
	 */
	buf->f_frsize = buf->f_bsize;
	buf->f_blocks = buf->f_bfree =  buf->f_bavail =
		1000ULL * 1024 * 1024 * 1024 / buf->f_frsize;
	buf->f_files = buf->f_ffree = 1000000000;
	return 0;
}
#else
static int sshfsm_statfs(const char *path, struct statfs *buf)
{
	(void) path;

	buf->f_namelen = 255;
	buf->f_bsize = sshfsm.blksize;
	buf->f_blocks = buf->f_bfree = buf->f_bavail =
		1000ULL * 1024 * 1024 * 1024 / buf->f_bsize;
	buf->f_files = buf->f_ffree = 1000000000;
	return 0;
}
#endif

#if FUSE_VERSION >= 25
static int sshfsm_create(const char *path, mode_t mode,
                         struct fuse_file_info *fi)
{
	struct table_entry *entry;
	GSList *idx_lst = NULL;
	int idx, err, i, ret = -INT_MAX;
	char *parent_dir;
	
	parent_dir = get_parent_dir(path);
	idx = table_tlb_lookup(parent_dir);
	if (idx != -1) {
		err = host_open_common(idx, path, mode, fi);
		ret = err > ret ? err : ret;
		if (!err) {
			g_free(parent_dir);
			goto out;
		}
	}
	
	entry = table_lookup_r(path);
	assert(entry != NULL);
	if (!entry)
		return -EIO;
	
	for (i = 0; i < entry->size; i++) {
		if ((idx = entry->idx_arr[i]) == -1)
			continue;

		err = host_open_common(idx, path, mode, fi);
		ret = err > ret ? err : ret;
		if (!err)
			break;
	}

  out:
	if (ret == 0) {
		table_tlb_update(path, idx);
		if (S_ISDIR(mode)) {
			idx_lst = g_slist_prepend(idx_lst, &idx);
			table_update(path, idx_lst);
		}
	}
	return ret;
}

static int sshfsm_ftruncate(const char *path, off_t size,
                           struct fuse_file_info *fi)
{
	int err;
	struct buffer buf;
	struct sshfsm_file *sf = get_sshfsm_file(fi);
	int idx = sf->host_idx;
	(void) path;
	
	if (!host_file_is_conn(idx, sf))
		return -EIO;

	sshfsm.host_arr[idx]->modifver ++;
	if (sshfsm.truncate_workaround)
		return host_truncate_workaround(idx, path, size, fi);

	buf_init(&buf, 0);
	buf_add_buf(&buf, &sf->handle);
	buf_add_uint32(&buf, SSH_FILEXFER_ATTR_SIZE);
	buf_add_uint64(&buf, size);
	err = sftp_request(idx, SSH_FXP_FSETSTAT, &buf, SSH_FXP_STATUS, NULL);
	buf_free(&buf);

	return err;
}

static int sshfsm_fgetattr(const char *path, 
			struct stat *stbuf, struct fuse_file_info *fi)
{
	int err;
	struct buffer buf, outbuf;
	struct sshfsm_file *sf = get_sshfsm_file(fi);
	int idx = sf->host_idx;
	(void) path;
	
	if (!host_file_is_conn(idx, sf))
		return -EIO;

	buf_init(&buf, 0);
	buf_add_buf(&buf, &sf->handle);
	err = sftp_request(idx, SSH_FXP_FSTAT, &buf, 
					   SSH_FXP_ATTRS, &outbuf);
	if (!err) {
		if (buf_get_attrs(&outbuf, stbuf, NULL) == -1)
			err = -EIO;
		buf_free(&outbuf);
	}
	buf_free(&buf);
	return err;
}
#endif

static int host_truncate_zero(int idx, const char *path)
{
	int err;
	struct fuse_file_info fi;

	fi.flags = O_WRONLY | O_TRUNC;
	err = host_open_common(idx, path, 0, &fi);
	if (!err)
		sshfsm_release(path, &fi);

	return err;
}

static size_t calc_buf_size(off_t size, off_t offset)
{
	return offset + sshfsm.max_read < size ? sshfsm.max_read : size - offset;
}

static int host_truncate_shrink(int idx, const char *path,
								  off_t size)
{
	int res;
	char *data;
	off_t offset;
	struct fuse_file_info fi;

	data = calloc(size, 1);
	if (!data)
		return -ENOMEM;

	fi.flags = O_RDONLY;
	res = host_open_common(idx, path, 0, &fi);
	if (res)
		goto out;

	for (offset = 0; offset < size; offset += res) {
		size_t bufsize = calc_buf_size(size, offset);
		res = sshfsm_read(path, data + offset, bufsize, 
						  offset, &fi);
		if (res <= 0)
			break;
	}
	sshfsm_release(path, &fi);
	if (res < 0)
		goto out;

	fi.flags = O_WRONLY | O_TRUNC;
	res = host_open_common(idx, path, 0, &fi);
	if (res)
		goto out;

	for (offset = 0; offset < size; offset += res) {
		size_t bufsize = calc_buf_size(size, offset);
		res = sshfsm_write(path, data + offset, bufsize, 
						   offset, &fi);
		if (res < 0)
			break;
	}
	if (res >= 0)
		res = sshfsm_flush(path, &fi);
	sshfsm_release(path, &fi);

out:
	free(data);
	return res;
}

static int host_truncate_extend(int idx, const char *path, 
					off_t size, struct fuse_file_info *fi)
{
	int res;
	char c = 0;
	struct fuse_file_info tmpfi;
	struct fuse_file_info *openfi = fi;
	if (!fi) {
		openfi = &tmpfi;
		openfi->flags = O_WRONLY;
		res = host_open_common(idx, path, 0, openfi);
		if (res)
			return res;
	}
	res = sshfsm_write(path, &c, 1, size - 1, openfi);
	if (res == 1)
		res = sshfsm_flush(path, openfi);
	if (!fi)
		sshfsm_release(path, openfi);

	return res;
}

/*
 * Work around broken sftp servers which don't handle
 * SSH_FILEXFER_ATTR_SIZE in SETSTAT request.
 *
 * If new size is zero, just open the file with O_TRUNC.
 *
 * If new size is smaller than current size, 
 * then copy file locally, then open/trunc and send it back.
 *
 * If new size is greater than current size, 
 * then write a zero byte to the new end of the file.
 */
static int host_truncate_workaround(int idx, const char *path,
					off_t size, struct fuse_file_info *fi)
{
	if (size == 0)
		return host_truncate_zero(idx, path);
	else {
		struct stat stbuf;
		int err;
		if (fi)
			err = sshfsm_fgetattr(path, &stbuf, fi);
		else
			err = host_getattr(idx, path, &stbuf);
		if (err)
			return err;
		if (stbuf.st_size == size)
			return 0;
		else if (stbuf.st_size > size)
			return host_truncate_shrink(idx, path, size);
		else
			return host_truncate_extend(idx, path, size, fi);
	}
}

static int processing_init(void)
{
	int i;
	struct host *hostp;
	
	pthread_mutex_init(&sshfsm.lock_host_arr, NULL);
	for (i = 0; i < sshfsm.host_num; i++) {
		hostp = sshfsm.host_arr[i];
		pthread_mutex_init(&hostp->lock, NULL);
		hostp->reqtab = g_hash_table_new(NULL, NULL);
		if (!hostp->reqtab) {
			fprintf(stderr, "failed to create request table\n");
			return -1;
		}
		pthread_mutex_init(&(hostp->lock_write), NULL);
		hostp->connver = hostp->processing_thread_started = 0;
		hostp->rank = i;
		hostp->fd = -1;
	}
	pthread_cond_init(&sshfsm.outstanding_cond, NULL);
	return 0;
}

static struct fuse_cache_operations sshfsm_oper = {
	.oper = {
#ifdef SSHFSM_USE_INIT
		.init       = sshfsm_init,
#endif
#ifdef SSHFSM_USE_DESTROY
		.destroy	= sshfsm_destroy,
#endif
		.getattr    = sshfsm_getattr,
		.readlink   = sshfsm_readlink,
		.mknod      = sshfsm_mknod,
		.mkdir      = sshfsm_mkdir,
		.symlink    = sshfsm_symlink,
		.unlink     = sshfsm_unlink,
		.rmdir      = sshfsm_rmdir,
		.rename     = sshfsm_rename,
		.chmod      = sshfsm_chmod,
		.chown      = sshfsm_chown,
		.truncate   = sshfsm_truncate,
		.utime      = sshfsm_utime,
		.open       = sshfsm_open,
		.flush      = sshfsm_flush,
		.fsync      = sshfsm_fsync,
		.release    = sshfsm_release,
		.read       = sshfsm_read,
		.write      = sshfsm_write,
		.statfs     = sshfsm_statfs,
#if FUSE_VERSION >= 25
		.create     = sshfsm_create,
		.ftruncate  = sshfsm_ftruncate,
		.fgetattr   = sshfsm_fgetattr,
#endif
	},
	.cache_getdir = sshfsm_getdir,
};

static void usage(const char *progname)
{
	fprintf(stderr,
"usage: \n"
"	%s [[user@]host:[dir]...] mountpoint [options]\n"
#ifdef SSHFSM_USE_CTRL
"	%s --umount mountpoint\n"
"	%s --add [[user@]host:[dir]...] mountpoint [options]\n"
"	%s --remove [[user@]host:[dir]...] mountpoint [options]\n"
#endif
"\n"
"general options:\n"
"    -o opt,[opt...]        mount options\n"
"    -h   --help            print help\n"
"    -V   --version         print version\n"
"\n"
"SSHFSM options:\n"
#ifdef SSHFSM_USE_CTRL
"    -u mountpoint          umount, equivalent to '--umount'\n"
"    -a                     add hosts, equivalent to '--add'\n"
"    -r                     remove hosts, equivalent to '--remove'\n"
#endif
"    -p PORT                equivalent to '-o port=PORT'\n"
"    -C                     equivalent to '-o compression=yes'\n"
"    -1                     equivalent to '-o ssh_protocol=1'\n"
"    -o reconnect           reconnect to server\n"
"    -o sshfsm_sync         synchronous writes\n"
"    -o no_readahead        synchronous reads (no speculative readahead)\n"
"    -o sshfsm_debug        print some debugging information\n"
"    -o cache=YESNO         enable caching {yes,no} (default: yes)\n"
"    -o cache_timeout=N     sets timeout for caches in seconds (default: 20)\n"
"    -o cache_X_timeout=N   sets timeout for {stat,dir,link} cache\n"
"    -o workaround=LIST     colon separated list of workarounds\n"
"             none             no workarounds enabled\n"
"             all              all workarounds enabled\n"
"             [no]rename       fix renaming to existing file (default: off)\n"
#ifdef SSH_NODELAY_WORKAROUND
"             [no]nodelay      set nodelay tcp flag in ssh (default: on)\n"
#endif
"             [no]nodelaysrv   set nodelay tcp flag in sshd (default: off)\n"
"             [no]truncate     fix truncate for old servers (default: off)\n"
"             [no]buflimit     fix buffer fillup bug in server (default: on)\n"
"    -o idmap=TYPE          user/group ID mapping, possible types are:\n"
"             none             no translation of the ID space (default)\n"
"             user             only translate UID of connecting user\n"
"    -o ssh_command=CMD     execute CMD instead of 'ssh'\n"
"    -o ssh_protocol=N      ssh protocol to use (default: 2)\n"
"    -o sftp_server=SERV    path to sftp server or subsystem (default: sftp)\n"
"    -o directport=PORT     directly connect to PORT bypassing ssh\n"
"    -o transform_symlinks  transform absolute symlinks to relative\n"
"    -o follow_symlinks     follow symlinks on the server\n"
"    -o no_check_root       don't check for existence of 'dir' on server\n"
"    -o SSHOPT=VAL          ssh options (see man ssh_config)\n"
"    -o conf_dir=NAME       config files directory (default: ~/.sshfsm)\n"
"    -o hosts_file=NAME     hosts list file (default: hosts)\n"
"    -o max_host=N			expected maximum host number (default: 1024)\n"
"    -o table_size          branch table size (default: 256)\n"
#ifdef SSHFSM_USE_CTRL
"    -o control=YESNO		enable control {yes,no} (default: yes)\n"
"\n", progname, progname, progname, progname
#else
"\n", progname
#endif
);}

static int is_ssh_opt(const char *arg)
{
	if (arg[0] != '-') {
		unsigned arglen = strlen(arg);
		const char **o;
		for (o = ssh_opts; *o; o++) {
			unsigned olen = strlen(*o);
			if (arglen > olen && arg[olen] == '=' &&
			    strncasecmp(arg, *o, olen) == 0)
				return 1;
		}
	}
	return 0;
}

static int sshfsm_fuse_main(struct fuse_args *args)
{
#if FUSE_VERSION >= 26
	return fuse_main(args->argc, args->argv, cache_init(&sshfsm_oper), NULL);
#else
	return fuse_main(args->argc, args->argv, cache_init(&sshfsm_oper));
#endif
}

static int sshfsm_opt_proc(void *data, const char *arg, int key,
                          struct fuse_args *outargs)
{
	char *tmp, *cp, *s;
	const char delimiters[] = ":";
	struct host *hostp;
	(void) data;

	switch (key) {
	case FUSE_OPT_KEY_OPT:
		if (is_ssh_opt(arg)) {
			tmp = g_strdup_printf("-o%s", arg);
			ssh_add_arg(tmp);
			g_free(tmp);
			return 0;
		}
		return 1;

	case FUSE_OPT_KEY_NONOPT:
		if (strchr(arg, ':')) {
			if (!sshfsm.host_arr) {
				sshfsm.host_arr = (struct host **) 
				sshfsm_calloc(sshfsm.host_num_max, sizeof(struct host *));
			}
			sshfsm.host_arr[sshfsm.host_num] = hostp = (struct host *) 
				sshfsm_malloc(sizeof(struct host));
			cp = strdup(arg);
			tmp = strtok(cp, delimiters);
			if (!tmp)
				goto out;
			else 
				hostp->host_name = strdup(tmp);
			
			tmp = strtok(NULL, delimiters);
			if (!tmp)
				hostp->base_path = strdup("");
			else {
				if (tmp[strlen(tmp) - 1] != '/') {
					s = (char *) 
						sshfsm_calloc(strlen(tmp) + 2,
									  sizeof(char));
					snprintf(s, strlen(tmp)+2, "%s/", tmp);
					hostp->base_path = s;
				}
				else
					hostp->base_path = strdup(tmp);
			}
			sshfsm.host_num++;
			do {
				sshfsm.host_arr_empty_idx = (sshfsm.host_arr_empty_idx + 1)
											% sshfsm.host_num_max;
			} while (sshfsm.host_arr[sshfsm.host_arr_empty_idx] != NULL);
		  out:
			free(cp);
			return 0;
		}
		return 1;

	case KEY_PORT:
		tmp = g_strdup_printf("-oPort=%s", arg + 2);
		ssh_add_arg(tmp);
		g_free(tmp);
		return 0;

	case KEY_COMPRESS:
		ssh_add_arg("-oCompression=yes");
		return 0;

	case KEY_HELP:
		usage(outargs->argv[0]);
		fuse_opt_add_arg(outargs, "-ho");
		sshfsm_fuse_main(outargs);
		exit(1);

	case KEY_VERSION:
		fprintf(stderr, "SSHFSM version %s\n", PACKAGE_VERSION);
#if FUSE_VERSION >= 25
		fuse_opt_add_arg(outargs, "--version");
		sshfsm_fuse_main(outargs);
#endif
		exit(0);
	
	case KEY_CTRL_UMOUNT:
		ctrl.is_ctrl = 1;
		ctrl.type = CTRL_TYPE_UMOUNT;
		return 0;
	
	case KEY_CTRL_ADD:
		ctrl.is_ctrl = 1;
		ctrl.type = CTRL_TYPE_ADD;
		return 0;
	
	case KEY_CTRL_REMOVE:
		ctrl.is_ctrl = 1;
		ctrl.type = CTRL_TYPE_REMOVE;
		return 0;
	
	default:
		fprintf(stderr, "internal error\n");
		abort();
	}
}

static int workaround_opt_proc(void *data, const char *arg, int key,
			       struct fuse_args *outargs)
{
	(void) data; (void) key; (void) outargs;
	fprintf(stderr, "unknown workaround: '%s'\n", arg);
	return -1;
}

int parse_workarounds(void)
{
	int res;
	char *argv[] = { "", "-o", sshfsm.workarounds, NULL };
	struct fuse_args args = FUSE_ARGS_INIT(3, argv);
	char *s = sshfsm.workarounds;
	if (!s)
		return 0;

	while ((s = strchr(s, ':')))
		*s = ',';

	res = fuse_opt_parse(&args, &sshfsm, workaround_opts,
			     workaround_opt_proc);
	fuse_opt_free_args(&args);

	return res;
}

#if FUSE_VERSION == 25
static int fuse_opt_insert_arg(struct fuse_args *args, int pos,
                               const char *arg)
{
	assert(pos <= args->argc);
	if (fuse_opt_add_arg(args, arg) == -1)
		return -1;

	if (pos != args->argc - 1) {
		char *newarg = args->argv[args->argc - 1];
		memmove(&args->argv[pos + 1], &args->argv[pos],
			sizeof(char *) * (args->argc - pos - 1));
		args->argv[pos] = newarg;
	}
	return 0;
}
#endif

static void ssh_insert_arg(const char *arg, int pos)
{
	if (fuse_opt_insert_arg(&sshfsm.ssh_args, pos, arg) == -1)
		_exit(1);
}

void check_large_read(struct fuse_args *args)
{
	struct utsname buf;
	int err = uname(&buf);
	if (!err && strcmp(buf.sysname, "Linux") == 0 &&
	    strncmp(buf.release, "2.4.", 4) == 0)
		fuse_opt_insert_arg(args, 1, "-olarge_read");
}

int main(int argc, char *argv[])
{
	int res;
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
	char *tmp;
	char *fsname;
	const char *sftp_server;
	int libver;

	g_thread_init(NULL);

	sshfsm.blksize = 4096;
	sshfsm.max_read = 65536;
	sshfsm.nodelay_workaround = 1;
	sshfsm.nodelaysrv_workaround = 0;
	sshfsm.rename_workaround = 0;
	sshfsm.truncate_workaround = 0;
	sshfsm.buflimit_workaround = 1;
	sshfsm.ssh_ver = 2;
	sshfsm.host_num_max = DEFAULT_MAX_HOST_NUM;
	sshfsm.host_num = 0;
	sshfsm.host_arr = NULL;
	sshfsm.progname = argv[0];
#ifdef	SSHFSM_USE_CTRL
	sshfsm.ctrl_on = 1;
	ctrl.is_ctrl = 0;
#endif
	ssh_add_arg("ssh");
	ssh_add_arg("-x");
	ssh_add_arg("-a");
	ssh_add_arg("-oClearAllForwardings=yes");
	
	if (fuse_opt_parse(&args, &sshfsm, sshfsm_opts, 
		sshfsm_opt_proc) == -1 || parse_workarounds() == -1)
		exit(1);
	
	if (ctrl.is_ctrl) {
		res = sshfsm_do_ctrl(&args);
		res == 0 ? exit(0) : exit(1);
	}

	if (table_parse_options(&args) == -1)
		exit(1);

	if (sshfsm.buflimit_workaround)
		/* Work around buggy sftp-server in OpenSSH.  
		 * Without this on a slow server a 10Mbyte buffer
		 * would fill up and the server would abort */
		sshfsm.max_outstanding_len = 8388608;
	else
		sshfsm.max_outstanding_len = ~0;
	
	sshfsm.conf_dir = sshfsm.conf_dir ? sshfsm.conf_dir : 
					  get_real_path(DEFAULT_CONF_PATH); 

	if (mkdir(sshfsm.conf_dir, S_IRWXU) == -1) {
		if (errno != EEXIST)
			fprintf(stderr, "make directory %s failed: %s\n",
					sshfsm.conf_dir, strerror(errno));
	};
	
	if (sshfsm.host_num == 0 || sshfsm.hosts_file != NULL) {
		if (sshfsm.hosts_file == NULL)
			sshfsm.hosts_file = strdup(DEFAULT_HOSTS_FILE_NAME);
		if (parse_hosts_file(sshfsm.hosts_file) == -1)
			exit(1);
	}

	if (sshfsm.host_num == 0) {
		fprintf(stderr, "missing host or hosts_file\n");
		fprintf(stderr, "see `%s -h' for usage\n", argv[0]);
		exit(1);
	}
	fsname = g_strdup_printf("[%dhosts]", sshfsm.host_num);

	if (sshfsm.ssh_command) {
		free(sshfsm.ssh_args.argv[0]);
		sshfsm.ssh_args.argv[0] = sshfsm.ssh_command;
	}

	if (fuse_parse_cmdline(&args, &(sshfsm.mountpoint), NULL, NULL) == -1)
		exit(1);

	if (!sshfsm.mountpoint) {
		fprintf(stderr, "sshfsm: missing mountpoint\n");
		fuse_opt_free_args(&args);
		exit(1);
	}
	fuse_opt_insert_arg(&args, 1, sshfsm.mountpoint);
	
	tmp = g_strdup_printf("-%i", sshfsm.ssh_ver);
	ssh_add_arg(tmp);
	g_free(tmp);

	if (sshfsm.sftp_server)
		sftp_server = sshfsm.sftp_server;
	else if (sshfsm.ssh_ver == 1)
		sftp_server = SFTP_SERVER_PATH;
	else
		sftp_server = "sftp";

	if (sshfsm.ssh_ver != 1 && 
		strchr(sftp_server, '/') == NULL)
		ssh_add_arg("-s");

	ssh_add_arg(sftp_server);
	free(sshfsm.sftp_server);
	
	/* Preserve a place for hostname argument in ssh_args 
	 * see start_ssh(1) */
	tmp = g_strdup_printf("hostname");
	ssh_insert_arg(tmp, 2);
	g_free(tmp);

	if (processing_init() == -1)
		exit(1);

	if (connect_remote_all() == -1)
		exit(1);

#ifndef SSHFSM_USE_INIT
	if (sshfsm.detect_uid)
		sftp_detect_uid_all();
#endif

	if (!sshfsm.no_check_root && sftp_check_root_all() == -1)
		exit(1);
	
	if (table_create(sshfsm.host_arr, sshfsm.host_num) == -1)
		exit(1);

	if (cache_parse_options(&args) == -1)
		exit(1);

	sshfsm.randseed = time(0);

	if (sshfsm.max_read > 65536)
		sshfsm.max_read = 65536;

	if (fuse_is_lib_option("ac_attr_timeout="))
		fuse_opt_insert_arg(&args, 1, "-oauto_cache,ac_attr_timeout=0");
	
	tmp = g_strdup_printf("-omax_read=%u", sshfsm.max_read);
	fuse_opt_insert_arg(&args, 1, tmp);
	g_free(tmp);
#if FUSE_VERSION >= 27
	libver = fuse_version();
	assert(libver >= 27);
	tmp = g_strdup_printf("-osubtype=sshfsm,fsname=%s", fsname);
#else
	tmp = g_strdup_printf("-ofsname=sshfsm#%s", fsname);
#endif
	fuse_opt_insert_arg(&args, 1, tmp);
	g_free(tmp);
	g_free(fsname);
	check_large_read(&args);
	res = sshfsm_fuse_main(&args);
	fuse_opt_free_args(&args);
	fuse_opt_free_args(&sshfsm.ssh_args);
	free(sshfsm.directport);
	return res;
}

/*******************************
  TODO:
  -- Recent
  * autoconf and automake
  * Evaluation
  *  - direcport
  *  - compare with Gfarm
  -- Long
  * Refine the code
  * libssh2 porting
*******************************/
